from typing import cast, Collection, Optional

from graphql.language.block_string import (
    is_printable_as_block_string,
    dedent_block_string_lines,
    print_block_string,
)


def join_lines(*args: str) -> str:
    return "\n".join(args)


def describe_dedent_block_string_lines():
    def _assert_dedent(lines: Collection[str], expected: Collection[str]) -> None:
        assert dedent_block_string_lines(lines) == expected

    def handles_empty_string():
        _assert_dedent([""], [])

    def does_not_dedent_first_line():
        _assert_dedent(["  a"], ["  a"])
        _assert_dedent([" a", "  b"], [" a", "b"])

    def removes_minimal_indentation_length():
        _assert_dedent(["", " a", "  b"], ["a", " b"])
        _assert_dedent(["", "  a", " b"], [" a", "b"])
        _assert_dedent(["", "  a", " b", "c"], ["  a", " b", "c"])

    def dedent_both_tab_and_space_as_single_character():
        _assert_dedent(["", "\ta", "          b"], ["a", "         b"])
        _assert_dedent(["", "\t a", "          b"], ["a", "        b"])
        _assert_dedent(["", " \t a", "          b"], ["a", "       b"])

    def removes_uniform_indentation_from_a_string():
        lines = ["", "    Hello,", "      World!", "", "    Yours,", "      GraphQL."]
        _assert_dedent(lines, ["Hello,", "  World!", "", "Yours,", "  GraphQL."])

    def removes_empty_leading_and_trailing_lines():
        lines = [
            "",
            "",
            "    Hello,",
            "      World!",
            "",
            "    Yours,",
            "      GraphQL.",
            "",
            "",
        ]
        _assert_dedent(lines, ["Hello,", "  World!", "", "Yours,", "  GraphQL."])

    def removes_blank_leading_and_trailing_lines():
        lines = [
            "  ",
            "        ",
            "    Hello,",
            "      World!",
            "",
            "    Yours,",
            "      GraphQL.",
            "        ",
            "  ",
        ]
        _assert_dedent(lines, ["Hello,", "  World!", "", "Yours,", "  GraphQL."])

    def retains_indentation_from_first_line():
        lines = ["    Hello,", "      World!", "", "    Yours,", "      GraphQL."]
        _assert_dedent(lines, ["    Hello,", "  World!", "", "Yours,", "  GraphQL."])

    def does_not_alter_trailing_spaces():
        lines = [
            "               ",
            "    Hello,     ",
            "      World!   ",
            "               ",
            "    Yours,     ",
            "      GraphQL. ",
            "               ",
        ]
        _assert_dedent(
            lines,
            [
                "Hello,     ",
                "  World!   ",
                "           ",
                "Yours,     ",
                "  GraphQL. ",
            ],
        )


def describe_is_printable_as_block_string():
    def _assert_printable(s: str) -> None:
        assert is_printable_as_block_string(s) is True, s

    def _assert_non_printable(s: str) -> None:
        assert is_printable_as_block_string(s) is False, s

    def accepts_valid_strings():
        _assert_printable("")
        _assert_printable(" a")
        _assert_printable('\t"\n"')
        _assert_non_printable('\t"\n \n\t"')

    def rejects_strings_with_only_whitespace():
        _assert_non_printable(" ")
        _assert_non_printable("\t")
        _assert_non_printable("\t ")
        _assert_non_printable(" \t")

    def rejects_strings_with_non_printable_characters():
        _assert_non_printable("\x00")
        _assert_non_printable("a\x00b")

    def rejects_strings_with_only_empty_lines():
        _assert_non_printable("\n")
        _assert_non_printable("\n\n")
        _assert_non_printable("\n\n\n")
        _assert_non_printable(" \n  \n")
        _assert_non_printable("\t\n\t\t\n")

    def rejects_strings_with_carriage_return():
        _assert_non_printable("\r")
        _assert_non_printable("\n\r")
        _assert_non_printable("\r\n")
        _assert_non_printable("a\rb")

    def rejects_strings_with_leading_empty_lines():
        _assert_non_printable("\na")
        _assert_non_printable(" \na")
        _assert_non_printable("\t\na")
        _assert_non_printable("\n\na")

    def rejects_strings_with_trailing_empty_lines():
        _assert_non_printable("a\n")
        _assert_non_printable("a\n ")
        _assert_non_printable("a\n\t")
        _assert_non_printable("a\n\n")

    def can_check_lazy_stings():
        class LazyString:
            def __init__(self, string: str) -> None:
                self.string = string

            def __str__(self) -> str:
                return self.string

        _assert_printable(cast(str, LazyString("")))
        _assert_non_printable(cast(str, LazyString(" ")))


def describe_print_block_string():
    def _assert_block_string(
        s: str, readable: str, minimize: Optional[str] = None
    ) -> None:
        assert print_block_string(s) == readable
        assert print_block_string(s, minimize=True) == minimize or readable

    def does_not_escape_characters():
        s = '" \\ / \b \f \n \r \t'
        _assert_block_string(s, f'"""\n{s}\n"""', f'"""\n{s}"""')

    def by_default_print_block_strings_as_single_line():
        s = "one liner"
        _assert_block_string(s, '"""one liner"""')

    def by_default_print_block_strings_ending_with_triple_quotation_as_multi_line():
        s = 'triple quotation """'
        _assert_block_string(
            s, '"""\ntriple quotation \\"""\n"""', '"""triple quotation \\""""""'
        )

    def correctly_prints_single_line_with_leading_space():
        s = "    space-led string"
        _assert_block_string(s, '"""    space-led string"""')

    def correctly_prints_single_line_with_leading_space_and_trailing_quotation():
        s = '    space-led value "quoted string"'
        _assert_block_string(s, '"""    space-led value "quoted string"\n"""')

    def correctly_prints_single_line_with_trailing_backslash():
        s = "backslash \\"
        _assert_block_string(s, '"""\nbackslash \\\n"""', '"""\nbackslash \\\n"""')

    def correctly_prints_multi_line_with_internal_indent():
        s = "no indent\n with indent"
        _assert_block_string(
            s, '"""\nno indent\n with indent\n"""', '"""\nno indent\n with indent"""'
        )

    def correctly_prints_string_with_a_first_line_indentation():
        s = join_lines("    first  ", "  line     ", "indentation", "     string")

        _assert_block_string(
            s,
            join_lines(
                '"""', "    first  ", "  line     ", "indentation", "     string", '"""'
            ),
            join_lines(
                '"""    first  ',
                "  line     ",
                "indentation",
                '     string"""',
            ),
        )

    def correctly_prints_lazy_stings():
        class LazyString:
            def __str__(self) -> str:
                return "lazy"

        _assert_block_string(cast(str, LazyString()), '"""lazy"""')



from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional

from graphql import graphql_sync
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)
from graphql.utilities import introspection_from_schema

ColorType = GraphQLEnumType("Color", values={"RED": 0, "GREEN": 1, "BLUE": 2})


class ColorTypeEnumValues(Enum):
    RED = 0
    GREEN = 1
    BLUE = 2


class Complex1:
    # noinspection PyMethodMayBeStatic
    some_random_object = datetime.now()


class Complex2:
    some_random_value = 123


complex1 = Complex1()
complex2 = Complex2()

ComplexEnum = GraphQLEnumType("Complex", {"ONE": complex1, "TWO": complex2})

ColorType2 = GraphQLEnumType("Color", ColorTypeEnumValues)

QueryType = GraphQLObjectType(
    "Query",
    {
        "colorEnum": GraphQLField(
            ColorType,
            args={
                "fromEnum": GraphQLArgument(ColorType),
                "fromInt": GraphQLArgument(GraphQLInt),
                "fromString": GraphQLArgument(GraphQLString),
            },
            resolve=lambda _source, info, **args: args.get("fromInt")
            or args.get("fromString")
            or args.get("fromEnum"),
        ),
        "colorInt": GraphQLField(
            GraphQLInt,
            args={
                "fromEnum": GraphQLArgument(ColorType),
                "fromInt": GraphQLArgument(GraphQLInt),
            },
            resolve=lambda _source, info, **args: args.get("fromEnum"),
        ),
        "complexEnum": GraphQLField(
            ComplexEnum,
            args={
                # Note: default_value is provided an *internal* representation for
                # Enums, rather than the string name.
                "fromEnum": GraphQLArgument(ComplexEnum, default_value=complex1),
                "provideGoodValue": GraphQLArgument(GraphQLBoolean),
                "provideBadValue": GraphQLArgument(GraphQLBoolean),
            },
            resolve=lambda _source, info, **args:
            # Note: this is one of the references of the internal values
            # which ComplexEnum allows.
            complex2 if args.get("provideGoodValue")
            # Note: similar object, but not the same *reference* as
            # complex2 above. Enum internal values require object equality.
            else Complex2() if args.get("provideBadValue") else args.get("fromEnum"),
        ),
    },
)

MutationType = GraphQLObjectType(
    "Mutation",
    {
        "favoriteEnum": GraphQLField(
            ColorType,
            args={"color": GraphQLArgument(ColorType)},
            resolve=lambda _source, info, color=None: color,
        )
    },
)

SubscriptionType = GraphQLObjectType(
    "Subscription",
    {
        "subscribeToEnum": GraphQLField(
            ColorType,
            args={"color": GraphQLArgument(ColorType)},
            resolve=lambda _source, info, color=None: color,
        )
    },
)

schema = GraphQLSchema(
    query=QueryType, mutation=MutationType, subscription=SubscriptionType
)


def execute_query(source: str, variable_values: Optional[Dict[str, Any]] = None):
    return graphql_sync(schema, source, variable_values=variable_values)


def describe_type_system_enum_values():
    def can_use_python_enums_instead_of_dicts():
        assert ColorType2.values == ColorType.values
        keys = [key for key in ColorType.values]
        keys2 = [key for key in ColorType2.values]
        assert keys2 == keys
        values = [value.value for value in ColorType.values.values()]
        values2 = [value.value for value in ColorType2.values.values()]
        assert values2 == values

    def accepts_enum_literals_as_input():
        result = execute_query("{ colorInt(fromEnum: GREEN) }")

        assert result == ({"colorInt": 1}, None)

    def enum_may_be_output_type():
        result = execute_query("{ colorEnum(fromInt: 1) }")

        assert result == ({"colorEnum": "GREEN"}, None)

    def enum_may_be_both_input_and_output_type():
        result = execute_query("{ colorEnum(fromEnum: GREEN) }")

        assert result == ({"colorEnum": "GREEN"}, None)

    def does_not_accept_string_literals():
        result = execute_query('{ colorEnum(fromEnum: "GREEN") }')

        assert result == (
            None,
            [
                {
                    "message": "Enum 'Color' cannot represent non-enum value:"
                    ' "GREEN".'
                    " Did you mean the enum value 'GREEN'?",
                    "locations": [(1, 23)],
                }
            ],
        )

    def does_not_accept_values_not_in_the_enum():
        result = execute_query("{ colorEnum(fromEnum: GREENISH) }")

        assert result == (
            None,
            [
                {
                    "message": "Value 'GREENISH' does not exist in 'Color' enum."
                    " Did you mean the enum value 'GREEN'?",
                    "locations": [(1, 23)],
                }
            ],
        )

    def does_not_accept_values_with_incorrect_casing():
        result = execute_query("{ colorEnum(fromEnum: green) }")

        assert result == (
            None,
            [
                {
                    "message": "Value 'green' does not exist in 'Color' enum."
                    " Did you mean the enum value 'GREEN' or 'RED'?",
                    "locations": [(1, 23)],
                }
            ],
        )

    def does_not_accept_incorrect_internal_value():
        result = execute_query('{ colorEnum(fromString: "GREEN") }')

        assert result == (
            {"colorEnum": None},
            [
                {
                    "message": "Enum 'Color' cannot represent value: 'GREEN'",
                    "locations": [(1, 3)],
                    "path": ["colorEnum"],
                }
            ],
        )

    def does_not_accept_internal_value_in_place_of_enum_literal():
        result = execute_query("{ colorEnum(fromEnum: 1) }")

        assert result == (
            None,
            [
                {
                    "message": "Enum 'Color' cannot represent non-enum value: 1.",
                    "locations": [(1, 23)],
                }
            ],
        )

    def does_not_accept_enum_literal_in_place_of_int():
        result = execute_query("{ colorEnum(fromInt: GREEN) }")

        assert result == (
            None,
            [
                {
                    "message": "Int cannot represent non-integer value: GREEN",
                    "locations": [(1, 22)],
                }
            ],
        )

    def accepts_json_string_as_enum_variable():
        doc = "query ($color: Color!) { colorEnum(fromEnum: $color) }"
        result = execute_query(doc, {"color": "BLUE"})

        assert result == ({"colorEnum": "BLUE"}, None)

    def accepts_enum_literals_as_input_arguments_to_mutations():
        doc = "mutation ($color: Color!) { favoriteEnum(color: $color) }"
        result = execute_query(doc, {"color": "GREEN"})

        assert result == ({"favoriteEnum": "GREEN"}, None)

    def accepts_enum_literals_as_input_arguments_to_subscriptions():
        doc = "subscription ($color: Color!) { subscribeToEnum(color: $color) }"
        result = execute_query(doc, {"color": "GREEN"})

        assert result == ({"subscribeToEnum": "GREEN"}, None)

    def does_not_accept_internal_value_as_enum_variable():
        doc = "query ($color: Color!) { colorEnum(fromEnum: $color) }"
        result = execute_query(doc, {"color": 2})

        assert result == (
            None,
            [
                {
                    "message": "Variable '$color' got invalid value 2;"
                    " Enum 'Color' cannot represent non-string value: 2.",
                    "locations": [(1, 8)],
                }
            ],
        )

    def does_not_accept_string_variables_as_enum_input():
        doc = "query ($color: String!) { colorEnum(fromEnum: $color) }"
        result = execute_query(doc, {"color": "BLUE"})

        assert result == (
            None,
            [
                {
                    "message": "Variable '$color' of type 'String!'"
                    " used in position expecting type 'Color'.",
                    "locations": [(1, 8), (1, 47)],
                }
            ],
        )

    def does_not_accept_internal_value_variable_as_enum_input():
        doc = "query ($color: Int!) { colorEnum(fromEnum: $color) }"
        result = execute_query(doc, {"color": 2})

        assert result == (
            None,
            [
                {
                    "message": "Variable '$color' of type 'Int!'"
                    " used in position expecting type 'Color'.",
                    "locations": [(1, 8), (1, 44)],
                }
            ],
        )

    def enum_value_may_have_an_internal_value_of_0():
        result = execute_query(
            """
            {
              colorEnum(fromEnum: RED)
              colorInt(fromEnum: RED)
            }
            """
        )

        assert result == ({"colorEnum": "RED", "colorInt": 0}, None)

    def enum_inputs_may_be_nullable():
        result = execute_query(
            """
            {
              colorEnum
              colorInt
            }
            """
        )

        assert result == ({"colorEnum": None, "colorInt": None}, None)

    def presents_a_values_property_for_complex_enums():
        values = ComplexEnum.values
        assert isinstance(values, dict)
        assert all(isinstance(value, GraphQLEnumValue) for value in values.values())
        assert {key: value.value for key, value in values.items()} == {
            "ONE": complex1,
            "TWO": complex2,
        }

    def may_be_internally_represented_with_complex_values():
        result = execute_query(
            """
            {
              first: complexEnum
              second: complexEnum(fromEnum: TWO)
              good: complexEnum(provideGoodValue: true)
              bad: complexEnum(provideBadValue: true)
            }
            """
        )

        assert result == (
            {"first": "ONE", "second": "TWO", "good": "TWO", "bad": None},
            [
                {
                    "message": "Enum 'Complex' cannot represent value:"
                    " <Complex2 instance>",
                    "locations": [(6, 15)],
                    "path": ["bad"],
                }
            ],
        )

    def can_be_introspected_without_error():
        introspection_from_schema(schema)



import weakref

from typing import cast, Tuple

from pytest import raises

from graphql.language import Source, SourceLocation

from ..utils import dedent


def describe_source():
    def accepts_body_and_name():
        source = Source("foo", "bar")
        assert source.body == "foo"
        assert source.name == "bar"

    def accepts_location_offset():
        location_offset = SourceLocation(2, 3)
        source = Source("", "", location_offset)
        assert source.location_offset is location_offset

    def accepts_tuple_as_location_offset():
        # noinspection PyTypeChecker
        source = Source("", "", (2, 3))  # type: ignore
        assert isinstance(source.location_offset, SourceLocation)
        assert source.location_offset == (2, 3)

    def uses_default_arguments():
        source = Source("")
        assert source.name == "GraphQL request"
        assert isinstance(source.location_offset, SourceLocation)
        assert source.location_offset == (1, 1)

    def can_get_location():
        body = dedent(
            """
            line 1
            line 2
            line 3
            """
        )
        source = Source(body)
        assert source.body == body
        location = source.get_location(body.find("2"))
        assert isinstance(location, SourceLocation)
        assert location == (2, 6)

    def can_be_stringified():
        source = Source("")
        assert str(source) == "<Source name='GraphQL request'>"

        source = Source("", "Custom source name")
        assert str(source) == "<Source name='Custom source name'>"

    def can_be_compared():
        source = Source("foo")
        assert source == source
        assert not source != source
        assert source == "foo"
        assert not source != "foo"
        same_source = Source("foo")
        assert source == same_source
        assert not source != same_source
        different_source = Source("bar")
        assert not source == different_source
        assert source != different_source
        assert not source == "bar"
        assert source != "bar"

    def can_create_weak_reference():
        source = Source("foo")
        ref = weakref.ref(source)
        assert ref() is source

    def can_create_custom_attribute():
        node = Source("foo")
        node.custom = "bar"  # type: ignore
        assert node.custom == "bar"  # type: ignore

    def rejects_invalid_location_offset():
        def create_source(location_offset: Tuple[int, int]) -> Source:
            return Source("", "", cast(SourceLocation, location_offset))

        with raises(TypeError):
            create_source(None)  # type: ignore
        with raises(TypeError):
            create_source(1)  # type: ignore
        with raises(TypeError):
            create_source((1,))  # type: ignore
        with raises(TypeError):
            create_source((1, 2, 3))  # type: ignore

        with raises(
            ValueError,
            match="line in location_offset is 1-indexed and must be positive\\.",
        ):
            create_source((0, 1))
        with raises(
            ValueError,
            match="line in location_offset is 1-indexed and must be positive\\.",
        ):
            create_source((-1, 1))

        with raises(
            ValueError,
            match="column in location_offset is 1-indexed and must be positive\\.",
        ):
            create_source((1, 0))
        with raises(
            ValueError,
            match="column in location_offset is 1-indexed and must be positive\\.",
        ):
            create_source((1, -1))



from graphql.type import (
    GraphQLSchema,
    GraphQLDeprecatedDirective,
    GraphQLIncludeDirective,
    GraphQLSkipDirective,
    GraphQLSpecifiedByDirective,
)
from graphql.utilities import (
    BreakingChangeType,
    DangerousChangeType,
    build_schema,
    find_breaking_changes,
    find_dangerous_changes,
)


def describe_find_breaking_changes():
    def should_detect_if_a_type_was_removed_or_not():
        old_schema = build_schema(
            """
            type Type1
            type Type2
            """
        )

        new_schema = build_schema(
            """
            type Type2
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (BreakingChangeType.TYPE_REMOVED, "Type1 was removed.")
        ]
        assert find_breaking_changes(old_schema, old_schema) == []

    def should_detect_if_a_standard_scalar_was_removed():
        old_schema = build_schema(
            """
            type Query {
              foo: Float
            }
            """
        )

        new_schema = build_schema(
            """
            type Query {
              foo: String
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.TYPE_REMOVED,
                "Standard scalar Float was removed"
                " because it is not referenced anymore.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Query.foo changed type from Float to String.",
            ),
        ]

    def should_detect_if_a_type_changed_its_type():
        old_schema = build_schema(
            """
            scalar TypeWasScalarBecomesEnum
            interface TypeWasInterfaceBecomesUnion
            type TypeWasObjectBecomesInputObject
            """
        )

        new_schema = build_schema(
            """
            enum TypeWasScalarBecomesEnum
            union TypeWasInterfaceBecomesUnion
           input TypeWasObjectBecomesInputObject
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.TYPE_CHANGED_KIND,
                "TypeWasScalarBecomesEnum changed from a Scalar type to an Enum type.",
            ),
            (
                BreakingChangeType.TYPE_CHANGED_KIND,
                "TypeWasInterfaceBecomesUnion changed"
                " from an Interface type to a Union type.",
            ),
            (
                BreakingChangeType.TYPE_CHANGED_KIND,
                "TypeWasObjectBecomesInputObject changed"
                " from an Object type to an Input type.",
            ),
        ]

    def should_detect_if_a_field_on_type_was_deleted_or_changed_type():
        old_schema = build_schema(
            """
            type TypeA
            type TypeB

            interface Type1 {
              field1: TypeA
              field2: String
              field3: String
              field4: TypeA
              field6: String
              field7: [String]
              field8: Int
              field9: Int!
              field10: [Int]!
              field11: Int
              field12: [Int]
              field13: [Int!]
              field14: [Int]
              field15: [[Int]]
              field16: Int!
              field17: [Int]
              field18: [[Int!]!]
            }
            """
        )

        new_schema = build_schema(
            """
            type TypeA
            type TypeB

            interface Type1 {
              field1: TypeA
              field3: Boolean
              field4: TypeB
              field5: String
              field6: [String]
              field7: String
              field8: Int!
              field9: Int
              field10: [Int]
              field11: [Int]!
              field12: [Int!]
              field13: [Int]
              field14: [[Int]]
              field15: [Int]
              field16: [Int]!
              field17: [Int]!
              field18: [[Int!]]
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (BreakingChangeType.FIELD_REMOVED, "Type1.field2 was removed."),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field3 changed type from String to Boolean.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field4 changed type from TypeA to TypeB.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field6 changed type from String to [String].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field7 changed type from [String] to String.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field9 changed type from Int! to Int.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field10 changed type from [Int]! to [Int].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field11 changed type from Int to [Int]!.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field13 changed type from [Int!] to [Int].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field14 changed type from [Int] to [[Int]].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field15 changed type from [[Int]] to [Int].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field16 changed type from Int! to [Int]!.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "Type1.field18 changed type from [[Int!]!] to [[Int!]].",
            ),
        ]

    def should_detect_if_fields_on_input_types_changed_kind_or_were_removed():
        old_schema = build_schema(
            """
            input InputType1 {
              field1: String
              field2: Boolean
              field3: [String]
              field4: String!
              field5: String
              field6: [Int]
              field7: [Int]!
              field8: Int
              field9: [Int]
              field10: [Int!]
              field11: [Int]
              field12: [[Int]]
              field13: Int!
              field14: [[Int]!]
              field15: [[Int]!]
            }
            """
        )

        new_schema = build_schema(
            """
            input InputType1 {
              field1: Int
              field3: String
              field4: String
              field5: String!
              field6: [Int]!
              field7: [Int]
              field8: [Int]!
              field9: [Int!]
              field10: [Int]
              field11: [[Int]]
              field12: [Int]
              field13: [Int]!
              field14: [[Int]]
              field15: [[Int!]!]
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (BreakingChangeType.FIELD_REMOVED, "InputType1.field2 was removed."),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field1 changed type from String to Int.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field3 changed type from [String] to String.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field5 changed type from String to String!.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field6 changed type from [Int] to [Int]!.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field8 changed type from Int to [Int]!.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field9 changed type from [Int] to [Int!].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field11 changed type from [Int] to [[Int]].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field12 changed type from [[Int]] to [Int].",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field13 changed type from Int! to [Int]!.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "InputType1.field15 changed type from [[Int]!] to [[Int!]!].",
            ),
        ]

    def should_detect_if_a_required_field_is_added_to_an_input_type():
        old_schema = build_schema(
            """
            input InputType1 {
              field1: String
            }
            """
        )

        new_schema = build_schema(
            """
            input InputType1 {
                field1: String
                requiredField: Int!
                optionalField1: Boolean
                optionalField2: Boolean! = false
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.REQUIRED_INPUT_FIELD_ADDED,
                "A required field requiredField on input type InputType1 was added.",
            )
        ]

    def should_detect_if_a_type_was_removed_from_a_union_type():
        old_schema = build_schema(
            """
            type Type1
            type Type2
            type Type3

            union UnionType1 = Type1 | Type2
            """
        )

        new_schema = build_schema(
            """
            type Type1
            type Type2
            type Type3

            union UnionType1 = Type1 | Type3
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.TYPE_REMOVED_FROM_UNION,
                "Type2 was removed from union type UnionType1.",
            )
        ]

    def should_detect_if_a_value_was_removed_from_an_enum_type():
        old_schema = build_schema(
            """
            enum EnumType1 {
              VALUE0
              VALUE1
              VALUE2
            }
            """
        )

        new_schema = build_schema(
            """
            enum EnumType1 {
              VALUE0
              VALUE2
              VALUE3
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.VALUE_REMOVED_FROM_ENUM,
                "VALUE1 was removed from enum type EnumType1.",
            )
        ]

    def should_detect_if_a_field_argument_was_removed():
        old_schema = build_schema(
            """
            interface Interface1 {
              field1(arg1: Boolean, objectArg: String): String
            }

            type Type1 {
              field1(name: String): String
            }
            """
        )

        new_schema = build_schema(
            """
            interface Interface1 {
              field1: String
            }

            type Type1 {
              field1: String
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (BreakingChangeType.ARG_REMOVED, "Interface1.field1 arg arg1 was removed."),
            (
                BreakingChangeType.ARG_REMOVED,
                "Interface1.field1 arg objectArg was removed.",
            ),
            (BreakingChangeType.ARG_REMOVED, "Type1.field1 arg name was removed."),
        ]

    def should_detect_if_a_field_argument_has_changed_type():
        old_schema = build_schema(
            """
            type Type1 {
              field1(
                arg1: String
                arg2: String
                arg3: [String]
                arg4: String
                arg5: String!
                arg6: String!
                arg7: [Int]!
                arg8: Int
                arg9: [Int]
                arg10: [Int!]
                arg11: [Int]
                arg12: [[Int]]
                arg13: Int!
                arg14: [[Int]!]
                arg15: [[Int]!]
              ): String
            }
            """
        )

        new_schema = build_schema(
            """
            type Type1 {
              field1(
                arg1: Int
                arg2: [String]
                arg3: String
                arg4: String!
                arg5: Int
                arg6: Int!
                arg7: [Int]
                arg8: [Int]!
                arg9: [Int!]
                arg10: [Int]
                arg11: [[Int]]
                arg12: [Int]
                arg13: [Int]!
                arg14: [[Int]]
                arg15: [[Int!]!]
               ): String
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg1 has changed type from String to Int.",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg2 has changed type from String to [String].",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg3 has changed type from [String] to String.",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg4 has changed type from String to String!.",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg5 has changed type from String! to Int.",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg6 has changed type from String! to Int!.",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg8 has changed type from Int to [Int]!.",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg9 has changed type from [Int] to [Int!].",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg11 has changed type from [Int] to [[Int]].",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg12 has changed type from [[Int]] to [Int].",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg13 has changed type from Int! to [Int]!.",
            ),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "Type1.field1 arg arg15 has changed type from [[Int]!] to [[Int!]!].",
            ),
        ]

    def should_detect_if_a_required_field_argument_was_added():
        old_schema = build_schema(
            """
            type Type1 {
              field1(arg1: String): String
            }
            """
        )

        new_schema = build_schema(
            """
            type Type1 {
              field1(
                arg1: String,
                newRequiredArg: String!
                newOptionalArg1: Int
                newOptionalArg2: Int! = 0
              ): String
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.REQUIRED_ARG_ADDED,
                "A required arg newRequiredArg on Type1.field1 was added.",
            )
        ]

    def should_not_flag_args_with_the_same_type_signature_as_breaking():
        old_schema = build_schema(
            """
            input InputType1 {
              field1: String
            }

            type Type1 {
              field1(arg1: Int!, arg2: InputType1): Int
            }
            """
        )

        new_schema = build_schema(
            """
            input InputType1 {
              field1: String
            }

            type Type1 {
              field1(arg1: Int!, arg2: InputType1): Int
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == []

    def should_consider_args_that_move_away_from_non_null_as_non_breaking():
        old_schema = build_schema(
            """
            type Type1 {
              field1(name: String!): String
            }
            """
        )

        new_schema = build_schema(
            """
            type Type1 {
              field1(name: String): String
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == []

    def should_detect_interfaces_removed_from_types():
        old_schema = build_schema(
            """
            interface Interface1

            type Type1 implements Interface1
            """
        )

        new_schema = build_schema(
            """
            interface Interface1

            type Type1
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.IMPLEMENTED_INTERFACE_REMOVED,
                "Type1 no longer implements interface Interface1.",
            )
        ]

    def should_detect_intrefaces_removed_from_interfaces():
        old_schema = build_schema(
            """
            interface Interface1

            interface Interface2 implements Interface1
            """
        )

        new_schema = build_schema(
            """
            interface Interface1

            interface Interface2
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.IMPLEMENTED_INTERFACE_REMOVED,
                "Interface2 no longer implements interface Interface1.",
            )
        ]

    def should_ignore_changes_in_order_of_interfaces():
        old_schema = build_schema(
            """
            interface FirstInterface
            interface SecondInterface

            type Type1 implements FirstInterface & SecondInterface
            """
        )

        new_schema = build_schema(
            """
            interface FirstInterface
            interface SecondInterface

            type Type1 implements SecondInterface & FirstInterface
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == []

    def should_detect_all_breaking_changes():
        old_schema = build_schema(
            """
            directive @DirectiveThatIsRemoved on FIELD_DEFINITION

            directive @DirectiveThatRemovesArg(arg1: String) on FIELD_DEFINITION

            directive @NonNullDirectiveAdded on FIELD_DEFINITION

            directive @DirectiveThatWasRepeatable repeatable on FIELD_DEFINITION

            directive @DirectiveName on FIELD_DEFINITION | QUERY

            type ArgThatChanges {
                field1(id: Float): String
            }

            enum EnumTypeThatLosesAValue {
                VALUE0
                VALUE1
                VALUE2
            }

            interface Interface1
            type TypeThatLoosesInterface1 implements Interface1

            type TypeInUnion1
            type TypeInUnion2
            union UnionTypeThatLosesAType = TypeInUnion1 | TypeInUnion2

            type TypeThatChangesType

            type TypeThatGetsRemoved

            interface TypeThatHasBreakingFieldChanges {
                field1: String
                field2: String
            }
            """
        )

        new_schema = build_schema(
            """
            directive @DirectiveThatRemovesArg on FIELD_DEFINITION

            directive @NonNullDirectiveAdded(arg1: Boolean!) on FIELD_DEFINITION

            directive @DirectiveThatWasRepeatable on FIELD_DEFINITION

            directive @DirectiveName on FIELD_DEFINITION

            type ArgThatChanges {
              field1(id: String): String
            }

            enum EnumTypeThatLosesAValue {
              VALUE1
              VALUE2
            }

            interface Interface1
            type TypeThatLoosesInterface1

            type TypeInUnion1
            type TypeInUnion2

            union UnionTypeThatLosesAType = TypeInUnion1

            interface TypeThatChangesType

            interface TypeThatHasBreakingFieldChanges {
              field2: Boolean
            }
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.TYPE_REMOVED,
                "Standard scalar Float was removed"
                " because it is not referenced anymore.",
            ),
            (BreakingChangeType.TYPE_REMOVED, "TypeThatGetsRemoved was removed."),
            (
                BreakingChangeType.ARG_CHANGED_KIND,
                "ArgThatChanges.field1 arg id has changed type from Float to String.",
            ),
            (
                BreakingChangeType.VALUE_REMOVED_FROM_ENUM,
                "VALUE0 was removed from enum type EnumTypeThatLosesAValue.",
            ),
            (
                BreakingChangeType.IMPLEMENTED_INTERFACE_REMOVED,
                "TypeThatLoosesInterface1 no longer implements interface Interface1.",
            ),
            (
                BreakingChangeType.TYPE_REMOVED_FROM_UNION,
                "TypeInUnion2 was removed from union type UnionTypeThatLosesAType.",
            ),
            (
                BreakingChangeType.TYPE_CHANGED_KIND,
                "TypeThatChangesType changed from an Object type to an"
                " Interface type.",
            ),
            (
                BreakingChangeType.FIELD_REMOVED,
                "TypeThatHasBreakingFieldChanges.field1 was removed.",
            ),
            (
                BreakingChangeType.FIELD_CHANGED_KIND,
                "TypeThatHasBreakingFieldChanges.field2 changed type"
                " from String to Boolean.",
            ),
            (
                BreakingChangeType.DIRECTIVE_REMOVED,
                "DirectiveThatIsRemoved was removed.",
            ),
            (
                BreakingChangeType.DIRECTIVE_ARG_REMOVED,
                "arg1 was removed from DirectiveThatRemovesArg.",
            ),
            (
                BreakingChangeType.REQUIRED_DIRECTIVE_ARG_ADDED,
                "A required arg arg1 on directive NonNullDirectiveAdded was added.",
            ),
            (
                BreakingChangeType.DIRECTIVE_REPEATABLE_REMOVED,
                "Repeatable flag was removed from DirectiveThatWasRepeatable.",
            ),
            (
                BreakingChangeType.DIRECTIVE_LOCATION_REMOVED,
                "QUERY was removed from DirectiveName.",
            ),
        ]

    def should_detect_if_a_directive_was_explicitly_removed():
        old_schema = build_schema(
            """
            directive @DirectiveThatIsRemoved on FIELD_DEFINITION
            directive @DirectiveThatStays on FIELD_DEFINITION
            """
        )

        new_schema = build_schema(
            """
            directive @DirectiveThatStays on FIELD_DEFINITION
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.DIRECTIVE_REMOVED,
                "DirectiveThatIsRemoved was removed.",
            )
        ]

    def should_detect_if_a_directive_was_implicitly_removed():
        old_schema = GraphQLSchema()

        new_schema = GraphQLSchema(
            directives=[
                GraphQLSkipDirective,
                GraphQLIncludeDirective,
                GraphQLSpecifiedByDirective,
            ]
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.DIRECTIVE_REMOVED,
                f"{GraphQLDeprecatedDirective.name} was removed.",
            )
        ]

    def should_detect_if_a_directive_argument_was_removed():
        old_schema = build_schema(
            """
            directive @DirectiveWithArg(arg1: String) on FIELD_DEFINITION
            """
        )

        new_schema = build_schema(
            """
            directive @DirectiveWithArg on FIELD_DEFINITION
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.DIRECTIVE_ARG_REMOVED,
                "arg1 was removed from DirectiveWithArg.",
            )
        ]

    def should_detect_if_an_optional_directive_argument_was_added():
        old_schema = build_schema(
            """
            directive @DirectiveName on FIELD_DEFINITION
            """
        )

        new_schema = build_schema(
            """
            directive @DirectiveName(
              newRequiredArg: String!
              newOptionalArg1: Int
              newOptionalArg2: Int! = 0
            ) on FIELD_DEFINITION
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.REQUIRED_DIRECTIVE_ARG_ADDED,
                "A required arg newRequiredArg on directive DirectiveName was added.",
            )
        ]

    def should_detect_removal_of_repeatable_flag():
        old_schema = build_schema(
            """
            directive @DirectiveName repeatable on OBJECT
            """
        )

        new_schema = build_schema(
            """
            directive @DirectiveName on OBJECT
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.DIRECTIVE_REPEATABLE_REMOVED,
                "Repeatable flag was removed from DirectiveName.",
            )
        ]

    def should_detect_locations_removed_from_a_directive():
        old_schema = build_schema(
            """
            directive @DirectiveName on FIELD_DEFINITION | QUERY
            """
        )

        new_schema = build_schema(
            """
            directive @DirectiveName on FIELD_DEFINITION
            """
        )

        assert find_breaking_changes(old_schema, new_schema) == [
            (
                BreakingChangeType.DIRECTIVE_LOCATION_REMOVED,
                "QUERY was removed from DirectiveName.",
            )
        ]


def describe_find_dangerous_changes():
    def should_detect_if_a_default_value_changed_on_an_argument():
        old_sdl = """
            input Input1 {
              innerInputArray: [Input2]
            }

            input Input2 {
              arrayField: [Int]
            }

            type Type1 {
              field1(
                withDefaultValue: String = "TO BE DELETED"
                stringArg: String = "test"
                emptyArray: [Int!] = []
                valueArray: [[String]] = [["a", "b"], ["c"]]
                complexObject: Input1 = {
                  innerInputArray: [{ arrayField: [1, 2, 3] }]
                }
              ): String
            }
            """

        old_schema = build_schema(old_sdl)
        copy_of_old_schema = build_schema(old_sdl)
        assert find_dangerous_changes(old_schema, copy_of_old_schema) == []

        new_schema = build_schema(
            """
            input Input1 {
              innerInputArray: [Input2]
            }

            input Input2 {
              arrayField: [Int]
            }

            type Type1 {
              field1(
                withDefaultValue: String
                stringArg: String = "Test"
                emptyArray: [Int!] = [7]
                valueArray: [[String]] = [["b", "a"], ["d"]]
                complexObject: Input1 = {
                  innerInputArray: [{ arrayField: [3, 2, 1] }]
                }
              ): String
            }
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                "Type1.field1 arg withDefaultValue defaultValue was removed.",
            ),
            (
                DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                "Type1.field1 arg stringArg has changed defaultValue"
                ' from "test" to "Test".',
            ),
            (
                DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                "Type1.field1 arg emptyArray has changed defaultValue from [] to [7].",
            ),
            (
                DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                "Type1.field1 arg valueArray has changed defaultValue"
                ' from [["a", "b"], ["c"]] to [["b", "a"], ["d"]].',
            ),
            (
                DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                "Type1.field1 arg complexObject has changed defaultValue"
                " from {innerInputArray: [{arrayField: [1, 2, 3]}]}"
                " to {innerInputArray: [{arrayField: [3, 2, 1]}]}.",
            ),
        ]

    def should_ignore_changes_in_field_order_of_default_value():
        old_schema = build_schema(
            """
            input Input1 {
              a: String
              b: String
              c: String
            }

            type Type1 {
              field1(
                arg1: Input1 = { a: "a", b: "b", c: "c" }
              ): String
            }
            """
        )

        new_schema = build_schema(
            """
           input Input1 {
             a: String
             b: String
             c: String
           }

           type Type1 {
             field1(
               arg1: Input1 = { c: "c", b: "b", a: "a" }
             ): String
           }
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == []

    def should_ignore_changes_in_field_definitions_order():
        old_schema = build_schema(
            """
            input Input1 {
              a: String
              b: String
              c: String
            }

            type Type1 {
              field1(
                arg1: Input1 = { a: "a", b: "b", c: "c" }
              ): String
            }
            """
        )

        new_schema = build_schema(
            """
            input Input1 {
              c: String
              b: String
              a: String
            }

            type Type1 {
              field1(
                arg1: Input1 = { a: "a", b: "b", c: "c" }
              ): String
            }
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == []

    def should_detect_if_a_value_was_added_to_an_enum_type():
        old_schema = build_schema(
            """
            enum EnumType1 {
              VALUE0
              VALUE1
            }
            """
        )

        new_schema = build_schema(
            """
            enum EnumType1 {
              VALUE0
              VALUE1
              VALUE2
            }
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.VALUE_ADDED_TO_ENUM,
                "VALUE2 was added to enum type EnumType1.",
            )
        ]

    def should_detect_interfaces_added_to_types():
        old_schema = build_schema(
            """
            interface OldInterface
            interface NewInterface

            type Type1 implements OldInterface
            """
        )

        new_schema = build_schema(
            """
            interface OldInterface
            interface NewInterface

            type Type1 implements OldInterface & NewInterface
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.IMPLEMENTED_INTERFACE_ADDED,
                "NewInterface added to interfaces implemented by Type1.",
            )
        ]

    def should_detect_interfaces_added_to_interfaces():
        old_schema = build_schema(
            """
            interface OldInterface
            interface NewInterface

            interface Interface1 implements OldInterface
            """
        )

        new_schema = build_schema(
            """
            interface OldInterface
            interface NewInterface

            interface Interface1 implements OldInterface & NewInterface
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.IMPLEMENTED_INTERFACE_ADDED,
                "NewInterface added to interfaces implemented by Interface1.",
            )
        ]

    def should_detect_if_a_type_was_added_to_a_union_type():
        old_schema = build_schema(
            """
            type Type1
            type Type2

            union UnionType1 = Type1
            """
        )

        new_schema = build_schema(
            """
            type Type1
            type Type2

            union UnionType1 = Type1 | Type2
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.TYPE_ADDED_TO_UNION,
                "Type2 was added to union type UnionType1.",
            )
        ]

    def should_detect_if_an_optional_field_was_added_to_an_input():
        old_schema = build_schema(
            """
            input InputType1 {
                field1: String
            }
            """
        )

        new_schema = build_schema(
            """
            input InputType1 {
              field1: String
              field2: Int
            }
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.OPTIONAL_INPUT_FIELD_ADDED,
                "An optional field field2 on input type InputType1 was added.",
            )
        ]

    def should_find_all_dangerous_changes():
        old_schema = build_schema(
            """
            enum EnumType1 {
              VALUE0
              VALUE1
            }

            type Type1 {
              field1(argThatChangesDefaultValue: String = "test"): String
            }

            interface Interface1
            type TypeThatGainsInterface1

            type TypeInUnion1
            union UnionTypeThatGainsAType = TypeInUnion1
            """
        )

        new_schema = build_schema(
            """
            enum EnumType1 {
              VALUE0
              VALUE1
              VALUE2
            }

            type Type1 {
              field1(argThatChangesDefaultValue: String = "Test"): String
            }

            interface Interface1
            type TypeThatGainsInterface1 implements Interface1

            type TypeInUnion1
            type TypeInUnion2
            union UnionTypeThatGainsAType = TypeInUnion1 | TypeInUnion2
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.VALUE_ADDED_TO_ENUM,
                "VALUE2 was added to enum type EnumType1.",
            ),
            (
                DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                "Type1.field1 arg argThatChangesDefaultValue has changed defaultValue"
                ' from "test" to "Test".',
            ),
            (
                DangerousChangeType.IMPLEMENTED_INTERFACE_ADDED,
                "Interface1 added to interfaces implemented"
                " by TypeThatGainsInterface1.",
            ),
            (
                DangerousChangeType.TYPE_ADDED_TO_UNION,
                "TypeInUnion2 was added to union type UnionTypeThatGainsAType.",
            ),
        ]

    def should_detect_if_an_optional_field_argument_was_added():
        old_schema = build_schema(
            """
            type Type1 {
              field1(arg1: String): String
            }
            """
        )

        new_schema = build_schema(
            """
            type Type1 {
              field1(arg1: String, arg2: String): String
            }
            """
        )

        assert find_dangerous_changes(old_schema, new_schema) == [
            (
                DangerousChangeType.OPTIONAL_ARG_ADDED,
                "An optional arg arg2 on Type1.field1 was added.",
            )
        ]



from typing import Collection, List
from sys import maxsize

__all__ = [
    "dedent_block_string_lines",
    "is_printable_as_block_string",
    "print_block_string",
]


def dedent_block_string_lines(lines: Collection[str]) -> List[str]:
    """Produce the value of a block string from its parsed raw value.

    This function works similar to CoffeeScript's block string,
    Python's docstring trim or Ruby's strip_heredoc.

    It implements the GraphQL spec's BlockStringValue() static algorithm.

    Note that this is very similar to Python's inspect.cleandoc() function.
    The difference is that the latter also expands tabs to spaces and
    removes whitespace at the beginning of the first line. Python also has
    textwrap.dedent() which uses a completely different algorithm.

    For internal use only.
    """
    common_indent = maxsize
    first_non_empty_line = None
    last_non_empty_line = -1

    for i, line in enumerate(lines):
        indent = leading_white_space(line)

        if indent == len(line):
            continue  # skip empty lines

        if first_non_empty_line is None:
            first_non_empty_line = i
        last_non_empty_line = i

        if i and indent < common_indent:
            common_indent = indent

    if first_non_empty_line is None:
        first_non_empty_line = 0

    return [  # Remove common indentation from all lines but first.
        line[common_indent:] if i else line for i, line in enumerate(lines)
    ][  # Remove leading and trailing blank lines.
        first_non_empty_line : last_non_empty_line + 1
    ]


def leading_white_space(s: str) -> int:
    i = 0
    for c in s:
        if c not in " \t":
            return i
        i += 1
    return i


def is_printable_as_block_string(value: str) -> bool:
    """Check whether the given string is printable as a block string.

    For internal use only.
    """
    if not isinstance(value, str):
        value = str(value)  # resolve lazy string proxy object

    if not value:
        return True  # emtpy string is printable

    is_empty_line = True
    has_indent = False
    has_common_indent = True
    seen_non_empty_line = False

    for c in value:
        if c == "\n":
            if is_empty_line and not seen_non_empty_line:
                return False  # has leading new line
            seen_non_empty_line = True
            is_empty_line = True
            has_indent = False
        elif c in " \t":
            has_indent = has_indent or is_empty_line
        elif c <= "\x0f":
            return False
        else:
            has_common_indent = has_common_indent and has_indent
            is_empty_line = False

    if is_empty_line:
        return False  # has trailing empty lines

    if has_common_indent and seen_non_empty_line:
        return False  # has internal indent

    return True


def print_block_string(value: str, minimize: bool = False) -> str:
    """Print a block string in the indented block form.

    Prints a block string in the indented block form by adding a leading and
    trailing blank line. However, if a block string starts with whitespace and
    is a single-line, adding a leading blank line would strip that whitespace.

    For internal use only.
    """
    if not isinstance(value, str):
        value = str(value)  # resolve lazy string proxy object

    escaped_value = value.replace('"""', '\\"""')

    # Expand a block string's raw value into independent lines.
    lines = escaped_value.splitlines() or [""]
    num_lines = len(lines)
    is_single_line = num_lines == 1

    # If common indentation is found,
    # we can fix some of those cases by adding a leading new line.
    force_leading_new_line = num_lines > 1 and all(
        not line or line[0] in " \t" for line in lines[1:]
    )

    # Trailing triple quotes just looks confusing but doesn't force trailing new line.
    has_trailing_triple_quotes = escaped_value.endswith('\\"""')

    # Trailing quote (single or double) or slash forces trailing new line
    has_trailing_quote = value.endswith('"') and not has_trailing_triple_quotes
    has_trailing_slash = value.endswith("\\")
    force_trailing_new_line = has_trailing_quote or has_trailing_slash

    print_as_multiple_lines = not minimize and (
        # add leading and trailing new lines only if it improves readability
        not is_single_line
        or len(value) > 70
        or force_trailing_new_line
        or force_leading_new_line
        or has_trailing_triple_quotes
    )

    # Format a multi-line block quote to account for leading space.
    skip_leading_new_line = is_single_line and value and value[0] in " \t"
    before = (
        "\n"
        if print_as_multiple_lines
        and not skip_leading_new_line
        or force_leading_new_line
        else ""
    )
    after = "\n" if print_as_multiple_lines or force_trailing_new_line else ""

    return f'"""{before}{escaped_value}{after}"""'



from graphql import (
    GraphQLSchema,
    GraphQLObjectType,
    GraphQLField,
    GraphQLString,
    graphql_sync,
)


user = GraphQLObjectType(
    name="User",
    fields={
        "id": GraphQLField(GraphQLString),
        "name": GraphQLField(GraphQLString),
    },
)


def resolve_user(obj, info):
    return {
        "id": "1",
        "name": "Sarah",
    }


schema = GraphQLSchema(
    query=GraphQLObjectType(
        name="Query",
        fields={
            "user": GraphQLField(
                user,
                resolve=resolve_user,
            )
        },
    )
)


def test_execute_basic_sync(benchmark):
    result = benchmark(lambda: graphql_sync(schema, "query { user { id, name }}"))
    assert not result.errors
    assert result.data == {
        "user": {
            "id": "1",
            "name": "Sarah",
        },
    }



from graphql import parse, DocumentNode

from ..fixtures import kitchen_sink_query  # noqa: F401


def test_parse_kitchen_sink(benchmark, kitchen_sink_query):  # noqa: F811
    query = benchmark(lambda: parse(kitchen_sink_query))
    assert isinstance(query, DocumentNode)



from enum import Enum
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from ..error import GraphQLError
from ..language import (
    EnumTypeDefinitionNode,
    EnumValueDefinitionNode,
    EnumTypeExtensionNode,
    EnumValueNode,
    FieldDefinitionNode,
    FieldNode,
    FragmentDefinitionNode,
    InputObjectTypeDefinitionNode,
    InputObjectTypeExtensionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    InterfaceTypeExtensionNode,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
    OperationDefinitionNode,
    ScalarTypeDefinitionNode,
    ScalarTypeExtensionNode,
    TypeDefinitionNode,
    TypeExtensionNode,
    UnionTypeDefinitionNode,
    UnionTypeExtensionNode,
    ValueNode,
    print_ast,
)
from ..pyutils import (
    AwaitableOrValue,
    Path,
    cached_property,
    did_you_mean,
    inspect,
    is_collection,
    is_description,
    suggestion_list,
    Undefined,
)
from ..utilities.value_from_ast_untyped import value_from_ast_untyped
from .assert_name import assert_name, assert_enum_value_name

try:
    from typing import TypedDict
except ImportError:  # Python < 3.8
    from typing_extensions import TypedDict

if TYPE_CHECKING:
    from .schema import GraphQLSchema  # noqa: F401

__all__ = [
    "is_type",
    "is_scalar_type",
    "is_object_type",
    "is_interface_type",
    "is_union_type",
    "is_enum_type",
    "is_input_object_type",
    "is_list_type",
    "is_non_null_type",
    "is_input_type",
    "is_output_type",
    "is_leaf_type",
    "is_composite_type",
    "is_abstract_type",
    "is_wrapping_type",
    "is_nullable_type",
    "is_named_type",
    "is_required_argument",
    "is_required_input_field",
    "assert_type",
    "assert_scalar_type",
    "assert_object_type",
    "assert_interface_type",
    "assert_union_type",
    "assert_enum_type",
    "assert_input_object_type",
    "assert_list_type",
    "assert_non_null_type",
    "assert_input_type",
    "assert_output_type",
    "assert_leaf_type",
    "assert_composite_type",
    "assert_abstract_type",
    "assert_wrapping_type",
    "assert_nullable_type",
    "assert_named_type",
    "get_nullable_type",
    "get_named_type",
    "resolve_thunk",
    "GraphQLAbstractType",
    "GraphQLArgument",
    "GraphQLArgumentKwargs",
    "GraphQLArgumentMap",
    "GraphQLCompositeType",
    "GraphQLEnumType",
    "GraphQLEnumTypeKwargs",
    "GraphQLEnumValue",
    "GraphQLEnumValueKwargs",
    "GraphQLEnumValueMap",
    "GraphQLField",
    "GraphQLFieldKwargs",
    "GraphQLFieldMap",
    "GraphQLFieldResolver",
    "GraphQLInputField",
    "GraphQLInputFieldKwargs",
    "GraphQLInputFieldMap",
    "GraphQLInputObjectType",
    "GraphQLInputObjectTypeKwargs",
    "GraphQLInputType",
    "GraphQLInterfaceType",
    "GraphQLInterfaceTypeKwargs",
    "GraphQLIsTypeOfFn",
    "GraphQLLeafType",
    "GraphQLList",
    "GraphQLNamedType",
    "GraphQLNamedTypeKwargs",
    "GraphQLNamedInputType",
    "GraphQLNamedOutputType",
    "GraphQLNullableType",
    "GraphQLNonNull",
    "GraphQLResolveInfo",
    "GraphQLScalarType",
    "GraphQLScalarTypeKwargs",
    "GraphQLScalarSerializer",
    "GraphQLScalarValueParser",
    "GraphQLScalarLiteralParser",
    "GraphQLObjectType",
    "GraphQLObjectTypeKwargs",
    "GraphQLOutputType",
    "GraphQLType",
    "GraphQLTypeResolver",
    "GraphQLUnionType",
    "GraphQLUnionTypeKwargs",
    "GraphQLWrappingType",
    "Thunk",
    "ThunkCollection",
    "ThunkMapping",
]


class GraphQLType:
    """Base class for all GraphQL types"""

    # Note: We don't use slots for GraphQLType objects because memory considerations
    # are not really important for the schema definition and it would make caching
    # properties slower or more complicated.


# There are predicates for each kind of GraphQL type.


def is_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLType)


def assert_type(type_: Any) -> GraphQLType:
    if not is_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL type.")
    return cast(GraphQLType, type_)


# These types wrap and modify other types

GT = TypeVar("GT", bound=GraphQLType)


class GraphQLWrappingType(GraphQLType, Generic[GT]):
    """Base class for all GraphQL wrapping types"""

    of_type: GT

    def __init__(self, type_: GT) -> None:
        if not is_type(type_):
            raise TypeError(
                f"Can only create a wrapper for a GraphQLType, but got: {type_}."
            )
        self.of_type = type_

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.of_type!r}>"


def is_wrapping_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLWrappingType)


def assert_wrapping_type(type_: Any) -> GraphQLWrappingType:
    if not is_wrapping_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL wrapping type.")
    return cast(GraphQLWrappingType, type_)


class GraphQLNamedTypeKwargs(TypedDict, total=False):
    name: str
    description: Optional[str]
    extensions: Dict[str, Any]
    # unfortunately, we cannot make the following more specific, because they are
    # used by subclasses with different node types and typed dicts cannot be refined
    ast_node: Optional[Any]
    extension_ast_nodes: Tuple[Any, ...]


class GraphQLNamedType(GraphQLType):
    """Base class for all GraphQL named types"""

    name: str
    description: Optional[str]
    extensions: Dict[str, Any]
    ast_node: Optional[TypeDefinitionNode]
    extension_ast_nodes: Tuple[TypeExtensionNode, ...]

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[TypeDefinitionNode] = None,
        extension_ast_nodes: Optional[Collection[TypeExtensionNode]] = None,
    ) -> None:
        assert_name(name)
        if description is not None and not is_description(description):
            raise TypeError("The description must be a string.")
        if extensions is None:
            extensions = {}
        elif not isinstance(extensions, dict) or not all(
            isinstance(key, str) for key in extensions
        ):
            raise TypeError(f"{name} extensions must be a dictionary with string keys.")
        if ast_node and not isinstance(ast_node, TypeDefinitionNode):
            raise TypeError(f"{name} AST node must be a TypeDefinitionNode.")
        if extension_ast_nodes:
            if not is_collection(extension_ast_nodes) or not all(
                isinstance(node, TypeExtensionNode) for node in extension_ast_nodes
            ):
                raise TypeError(
                    f"{name} extension AST nodes must be specified"
                    " as a collection of TypeExtensionNode instances."
                )
            if not isinstance(extension_ast_nodes, tuple):
                extension_ast_nodes = tuple(extension_ast_nodes)
        else:
            extension_ast_nodes = ()
        self.name = name
        self.description = description
        self.extensions = extensions
        self.ast_node = ast_node
        self.extension_ast_nodes = extension_ast_nodes

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name!r}>"

    def __str__(self) -> str:
        return self.name

    def to_kwargs(self) -> GraphQLNamedTypeKwargs:
        return GraphQLNamedTypeKwargs(
            name=self.name,
            description=self.description,
            extensions=self.extensions,
            ast_node=self.ast_node,
            extension_ast_nodes=self.extension_ast_nodes,
        )

    def __copy__(self) -> "GraphQLNamedType":  # pragma: no cover
        return self.__class__(**self.to_kwargs())


T = TypeVar("T")

ThunkCollection = Union[Callable[[], Collection[T]], Collection[T]]
ThunkMapping = Union[Callable[[], Mapping[str, T]], Mapping[str, T]]
Thunk = Union[Callable[[], T], T]


def resolve_thunk(thunk: Thunk[T]) -> T:
    """Resolve the given thunk.

    Used while defining GraphQL types to allow for circular references in otherwise
    immutable type definitions.
    """
    return thunk() if callable(thunk) else thunk


GraphQLScalarSerializer = Callable[[Any], Any]
GraphQLScalarValueParser = Callable[[Any], Any]
GraphQLScalarLiteralParser = Callable[[ValueNode, Optional[Dict[str, Any]]], Any]


class GraphQLScalarTypeKwargs(GraphQLNamedTypeKwargs, total=False):
    serialize: Optional[GraphQLScalarSerializer]
    parse_value: Optional[GraphQLScalarValueParser]
    parse_literal: Optional[GraphQLScalarLiteralParser]
    specified_by_url: Optional[str]


class GraphQLScalarType(GraphQLNamedType):
    """Scalar Type Definition

    The leaf values of any request and input values to arguments are Scalars (or Enums)
    and are defined with a name and a series of functions used to parse input from ast
    or variables and to ensure validity.

    If a type's serialize function returns ``None``, then an error will be raised and a
    ``None`` value will be returned in the response. It is always better to validate.

    Example::

        def serialize_odd(value: Any) -> int:
            try:
                value = int(value)
            except ValueError:
                raise GraphQLError(
                    f"Scalar 'Odd' cannot represent '{value}'"
                    " since it is not an integer.")
            if not value % 2:
                raise GraphQLError(
                    f"Scalar 'Odd' cannot represent '{value}' since it is even.")
            return value

        odd_type = GraphQLScalarType('Odd', serialize=serialize_odd)

    """

    specified_by_url: Optional[str]
    ast_node: Optional[ScalarTypeDefinitionNode]
    extension_ast_nodes: Tuple[ScalarTypeExtensionNode, ...]

    def __init__(
        self,
        name: str,
        serialize: Optional[GraphQLScalarSerializer] = None,
        parse_value: Optional[GraphQLScalarValueParser] = None,
        parse_literal: Optional[GraphQLScalarLiteralParser] = None,
        description: Optional[str] = None,
        specified_by_url: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[ScalarTypeDefinitionNode] = None,
        extension_ast_nodes: Optional[Collection[ScalarTypeExtensionNode]] = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        if specified_by_url is not None and not isinstance(specified_by_url, str):
            raise TypeError(
                f"{name} must provide 'specified_by_url' as a string,"
                f" but got: {inspect(specified_by_url)}."
            )
        if serialize is not None and not callable(serialize):
            raise TypeError(
                f"{name} must provide 'serialize' as a function."
                " If this custom Scalar is also used as an input type,"
                " ensure 'parse_value' and 'parse_literal' functions"
                " are also provided."
            )
        if parse_literal is not None and (
            not callable(parse_literal)
            or (parse_value is None or not callable(parse_value))
        ):
            raise TypeError(
                f"{name} must provide"
                " both 'parse_value' and 'parse_literal' as functions."
            )
        if ast_node and not isinstance(ast_node, ScalarTypeDefinitionNode):
            raise TypeError(f"{name} AST node must be a ScalarTypeDefinitionNode.")
        if extension_ast_nodes and not all(
            isinstance(node, ScalarTypeExtensionNode) for node in extension_ast_nodes
        ):
            raise TypeError(
                f"{name} extension AST nodes must be specified"
                " as a collection of ScalarTypeExtensionNode instances."
            )
        if serialize is not None:
            self.serialize = serialize  # type: ignore
        if parse_value is not None:
            self.parse_value = parse_value  # type: ignore
        if parse_literal is not None:
            self.parse_literal = parse_literal  # type: ignore
        self.specified_by_url = specified_by_url

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name!r}>"

    def __str__(self) -> str:
        return self.name

    @staticmethod
    def serialize(value: Any) -> Any:
        """Serializes an internal value to include in a response.

        This default method just passes the value through and should be replaced
        with a more specific version when creating a scalar type.
        """
        return value

    @staticmethod
    def parse_value(value: Any) -> Any:
        """Parses an externally provided value to use as an input.

        This default method just passes the value through and should be replaced
        with a more specific version when creating a scalar type.
        """
        return value

    def parse_literal(
        self, node: ValueNode, variables: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Parses an externally provided literal value to use as an input.

        This default method uses the parse_value method and should be replaced
        with a more specific version when creating a scalar type.
        """
        return self.parse_value(value_from_ast_untyped(node, variables))

    def to_kwargs(self) -> GraphQLScalarTypeKwargs:
        # noinspection PyArgumentList
        return GraphQLScalarTypeKwargs(  # type: ignore
            super().to_kwargs(),
            serialize=None
            if self.serialize is GraphQLScalarType.serialize
            else self.serialize,
            parse_value=None
            if self.parse_value is GraphQLScalarType.parse_value
            else self.parse_value,
            parse_literal=None
            if getattr(self.parse_literal, "__func__", None)
            is GraphQLScalarType.parse_literal
            else self.parse_literal,
            specified_by_url=self.specified_by_url,
        )

    def __copy__(self) -> "GraphQLScalarType":  # pragma: no cover
        return self.__class__(**self.to_kwargs())


def is_scalar_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLScalarType)


def assert_scalar_type(type_: Any) -> GraphQLScalarType:
    if not is_scalar_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL Scalar type.")
    return cast(GraphQLScalarType, type_)


GraphQLArgumentMap = Dict[str, "GraphQLArgument"]


class GraphQLFieldKwargs(TypedDict, total=False):
    type_: "GraphQLOutputType"
    args: Optional[GraphQLArgumentMap]
    resolve: Optional["GraphQLFieldResolver"]
    subscribe: Optional["GraphQLFieldResolver"]
    description: Optional[str]
    deprecation_reason: Optional[str]
    extensions: Dict[str, Any]
    ast_node: Optional[FieldDefinitionNode]


class GraphQLField:
    """Definition of a GraphQL field"""

    type: "GraphQLOutputType"
    args: GraphQLArgumentMap
    resolve: Optional["GraphQLFieldResolver"]
    subscribe: Optional["GraphQLFieldResolver"]
    description: Optional[str]
    deprecation_reason: Optional[str]
    extensions: Dict[str, Any]
    ast_node: Optional[FieldDefinitionNode]

    def __init__(
        self,
        type_: "GraphQLOutputType",
        args: Optional[GraphQLArgumentMap] = None,
        resolve: Optional["GraphQLFieldResolver"] = None,
        subscribe: Optional["GraphQLFieldResolver"] = None,
        description: Optional[str] = None,
        deprecation_reason: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[FieldDefinitionNode] = None,
    ) -> None:
        if not is_output_type(type_):
            raise TypeError("Field type must be an output type.")
        if args is None:
            args = {}
        elif not isinstance(args, dict):
            raise TypeError("Field args must be a dict with argument names as keys.")
        elif not all(
            isinstance(value, GraphQLArgument) or is_input_type(value)
            for value in args.values()
        ):
            raise TypeError(
                "Field args must be GraphQLArguments or input type objects."
            )
        else:
            args = {
                assert_name(name): value
                if isinstance(value, GraphQLArgument)
                else GraphQLArgument(cast(GraphQLInputType, value))
                for name, value in args.items()
            }
        if resolve is not None and not callable(resolve):
            raise TypeError(
                "Field resolver must be a function if provided, "
                f" but got: {inspect(resolve)}."
            )
        if description is not None and not is_description(description):
            raise TypeError("The description must be a string.")
        if deprecation_reason is not None and not is_description(deprecation_reason):
            raise TypeError("The deprecation reason must be a string.")
        if extensions is None:
            extensions = {}
        elif not isinstance(extensions, dict) or not all(
            isinstance(key, str) for key in extensions
        ):
            raise TypeError("Field extensions must be a dictionary with string keys.")
        if ast_node and not isinstance(ast_node, FieldDefinitionNode):
            raise TypeError("Field AST node must be a FieldDefinitionNode.")
        self.type = type_
        self.args = args or {}
        self.resolve = resolve
        self.subscribe = subscribe
        self.description = description
        self.deprecation_reason = deprecation_reason
        self.extensions = extensions
        self.ast_node = ast_node

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.type!r}>"

    def __str__(self) -> str:
        return f"Field: {self.type}"

    def __eq__(self, other: Any) -> bool:
        return self is other or (
            isinstance(other, GraphQLField)
            and self.type == other.type
            and self.args == other.args
            and self.resolve == other.resolve
            and self.description == other.description
            and self.deprecation_reason == other.deprecation_reason
            and self.extensions == other.extensions
        )

    def to_kwargs(self) -> GraphQLFieldKwargs:
        return GraphQLFieldKwargs(
            type_=self.type,
            args=self.args.copy() if self.args else None,
            resolve=self.resolve,
            subscribe=self.subscribe,
            deprecation_reason=self.deprecation_reason,
            description=self.description,
            extensions=self.extensions,
            ast_node=self.ast_node,
        )

    def __copy__(self) -> "GraphQLField":  # pragma: no cover
        return self.__class__(**self.to_kwargs())


class GraphQLResolveInfo(NamedTuple):
    """Collection of information passed to the resolvers.

    This is always passed as the first argument to the resolvers.

    Note that contrary to the JavaScript implementation, the context (commonly used to
    represent an authenticated user, or request-specific caches) is included here and
    not passed as an additional argument.
    """

    field_name: str
    field_nodes: List[FieldNode]
    return_type: "GraphQLOutputType"
    parent_type: "GraphQLObjectType"
    path: Path
    schema: "GraphQLSchema"
    fragments: Dict[str, FragmentDefinitionNode]
    root_value: Any
    operation: OperationDefinitionNode
    variable_values: Dict[str, Any]
    context: Any
    is_awaitable: Callable[[Any], bool]


# Note: Contrary to the Javascript implementation of GraphQLFieldResolver,
# the context is passed as part of the GraphQLResolveInfo and any arguments
# are passed individually as keyword arguments.
GraphQLFieldResolverWithoutArgs = Callable[[Any, GraphQLResolveInfo], Any]
# Unfortunately there is currently no syntax to indicate optional or keyword
# arguments in Python, so we also allow any other Callable as a workaround:
GraphQLFieldResolver = Callable[..., Any]

# Note: Contrary to the Javascript implementation of GraphQLTypeResolver,
# the context is passed as part of the GraphQLResolveInfo:
GraphQLTypeResolver = Callable[
    [Any, GraphQLResolveInfo, "GraphQLAbstractType"],
    AwaitableOrValue[Optional[str]],
]

# Note: Contrary to the Javascript implementation of GraphQLIsTypeOfFn,
# the context is passed as part of the GraphQLResolveInfo:
GraphQLIsTypeOfFn = Callable[[Any, GraphQLResolveInfo], AwaitableOrValue[bool]]

GraphQLFieldMap = Dict[str, GraphQLField]


class GraphQLArgumentKwargs(TypedDict, total=False):
    type_: "GraphQLInputType"
    default_value: Any
    description: Optional[str]
    deprecation_reason: Optional[str]
    out_name: Optional[str]
    extensions: Dict[str, Any]
    ast_node: Optional[InputValueDefinitionNode]


class GraphQLArgument:
    """Definition of a GraphQL argument"""

    type: "GraphQLInputType"
    default_value: Any
    description: Optional[str]
    deprecation_reason: Optional[str]
    out_name: Optional[str]  # for transforming names (extension of GraphQL.js)
    extensions: Dict[str, Any]
    ast_node: Optional[InputValueDefinitionNode]

    def __init__(
        self,
        type_: "GraphQLInputType",
        default_value: Any = Undefined,
        description: Optional[str] = None,
        deprecation_reason: Optional[str] = None,
        out_name: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[InputValueDefinitionNode] = None,
    ) -> None:
        if not is_input_type(type_):
            raise TypeError("Argument type must be a GraphQL input type.")
        if description is not None and not is_description(description):
            raise TypeError("Argument description must be a string.")
        if deprecation_reason is not None and not is_description(deprecation_reason):
            raise TypeError("Argument deprecation reason must be a string.")
        if out_name is not None and not isinstance(out_name, str):
            raise TypeError("Argument out name must be a string.")
        if extensions is None:
            extensions = {}
        elif not isinstance(extensions, dict) or not all(
            isinstance(key, str) for key in extensions
        ):
            raise TypeError(
                "Argument extensions must be a dictionary with string keys."
            )
        if ast_node and not isinstance(ast_node, InputValueDefinitionNode):
            raise TypeError("Argument AST node must be an InputValueDefinitionNode.")
        self.type = type_
        self.default_value = default_value
        self.description = description
        self.deprecation_reason = deprecation_reason
        self.out_name = out_name
        self.extensions = extensions
        self.ast_node = ast_node

    def __eq__(self, other: Any) -> bool:
        return self is other or (
            isinstance(other, GraphQLArgument)
            and self.type == other.type
            and self.default_value == other.default_value
            and self.description == other.description
            and self.deprecation_reason == other.deprecation_reason
            and self.out_name == other.out_name
            and self.extensions == other.extensions
        )

    def to_kwargs(self) -> GraphQLArgumentKwargs:
        return GraphQLArgumentKwargs(
            type_=self.type,
            default_value=self.default_value,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            out_name=self.out_name,
            extensions=self.extensions,
            ast_node=self.ast_node,
        )

    def __copy__(self) -> "GraphQLArgument":  # pragma: no cover
        return self.__class__(**self.to_kwargs())


def is_required_argument(arg: GraphQLArgument) -> bool:
    return is_non_null_type(arg.type) and arg.default_value is Undefined


class GraphQLObjectTypeKwargs(GraphQLNamedTypeKwargs, total=False):

    fields: GraphQLFieldMap
    interfaces: Tuple["GraphQLInterfaceType", ...]
    is_type_of: Optional[GraphQLIsTypeOfFn]


class GraphQLObjectType(GraphQLNamedType):
    """Object Type Definition

    Almost all the GraphQL types you define will be object types. Object types have
    a name, but most importantly describe their fields.

    Example::

        AddressType = GraphQLObjectType('Address', {
            'street': GraphQLField(GraphQLString),
            'number': GraphQLField(GraphQLInt),
            'formatted': GraphQLField(GraphQLString,
                lambda obj, info, **args: f'{obj.number} {obj.street}')
        })

    When two types need to refer to each other, or a type needs to refer to itself in
    a field, you can use a lambda function with no arguments (a so-called "thunk")
    to supply the fields lazily.

    Example::

        PersonType = GraphQLObjectType('Person', lambda: {
            'name': GraphQLField(GraphQLString),
            'bestFriend': GraphQLField(PersonType)
        })

    """

    is_type_of: Optional[GraphQLIsTypeOfFn]
    ast_node: Optional[ObjectTypeDefinitionNode]
    extension_ast_nodes: Tuple[ObjectTypeExtensionNode, ...]

    def __init__(
        self,
        name: str,
        fields: ThunkMapping[GraphQLField],
        interfaces: Optional[ThunkCollection["GraphQLInterfaceType"]] = None,
        is_type_of: Optional[GraphQLIsTypeOfFn] = None,
        extensions: Optional[Dict[str, Any]] = None,
        description: Optional[str] = None,
        ast_node: Optional[ObjectTypeDefinitionNode] = None,
        extension_ast_nodes: Optional[Collection[ObjectTypeExtensionNode]] = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        if is_type_of is not None and not callable(is_type_of):
            raise TypeError(
                f"{name} must provide 'is_type_of' as a function,"
                f" but got: {inspect(is_type_of)}."
            )
        if ast_node and not isinstance(ast_node, ObjectTypeDefinitionNode):
            raise TypeError(f"{name} AST node must be an ObjectTypeDefinitionNode.")
        if extension_ast_nodes and not all(
            isinstance(node, ObjectTypeExtensionNode) for node in extension_ast_nodes
        ):
            raise TypeError(
                f"{name} extension AST nodes must be specified"
                " as a collection of ObjectTypeExtensionNode instances."
            )
        self._fields = fields
        self._interfaces = interfaces
        self.is_type_of = is_type_of

    def to_kwargs(self) -> GraphQLObjectTypeKwargs:
        # noinspection PyArgumentList
        return GraphQLObjectTypeKwargs(  # type: ignore
            super().to_kwargs(),
            fields=self.fields.copy(),
            interfaces=self.interfaces,
            is_type_of=self.is_type_of,
        )

    def __copy__(self) -> "GraphQLObjectType":  # pragma: no cover
        return self.__class__(**self.to_kwargs())

    @cached_property
    def fields(self) -> GraphQLFieldMap:
        """Get provided fields, wrapping them as GraphQLFields if needed."""
        try:
            fields = resolve_thunk(self._fields)
        except Exception as error:
            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
            raise cls(f"{self.name} fields cannot be resolved. {error}") from error
        if not isinstance(fields, Mapping) or not all(
            isinstance(key, str) for key in fields
        ):
            raise TypeError(
                f"{self.name} fields must be specified"
                " as a mapping with field names as keys."
            )
        if not all(
            isinstance(value, GraphQLField) or is_output_type(value)
            for value in fields.values()
        ):
            raise TypeError(
                f"{self.name} fields must be GraphQLField or output type objects."
            )
        return {
            assert_name(name): value
            if isinstance(value, GraphQLField)
            else GraphQLField(value)  # type: ignore
            for name, value in fields.items()
        }

    @cached_property
    def interfaces(self) -> Tuple["GraphQLInterfaceType", ...]:
        """Get provided interfaces."""
        try:
            interfaces: Collection["GraphQLInterfaceType"] = resolve_thunk(
                self._interfaces  # type: ignore
            )
        except Exception as error:
            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
            raise cls(f"{self.name} interfaces cannot be resolved. {error}") from error
        if interfaces is None:
            interfaces = ()
        elif not is_collection(interfaces) or not all(
            isinstance(value, GraphQLInterfaceType) for value in interfaces
        ):
            raise TypeError(
                f"{self.name} interfaces must be specified"
                " as a collection of GraphQLInterfaceType instances."
            )
        return tuple(interfaces)


def is_object_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLObjectType)


def assert_object_type(type_: Any) -> GraphQLObjectType:
    if not is_object_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL Object type.")
    return cast(GraphQLObjectType, type_)


class GraphQLInterfaceTypeKwargs(GraphQLNamedTypeKwargs, total=False):
    fields: GraphQLFieldMap
    interfaces: Tuple["GraphQLInterfaceType", ...]
    resolve_type: Optional[GraphQLTypeResolver]


class GraphQLInterfaceType(GraphQLNamedType):
    """Interface Type Definition

    When a field can return one of a heterogeneous set of types, an Interface type
    is used to describe what types are possible, what fields are in common across
    all types, as well as a function to determine which type is actually used when
    the field is resolved.

    Example::

        EntityType = GraphQLInterfaceType('Entity', {
                'name': GraphQLField(GraphQLString),
            })
    """

    resolve_type: Optional[GraphQLTypeResolver]
    ast_node: Optional[InterfaceTypeDefinitionNode]
    extension_ast_nodes: Tuple[InterfaceTypeExtensionNode, ...]

    def __init__(
        self,
        name: str,
        fields: ThunkMapping[GraphQLField],
        interfaces: Optional[ThunkCollection["GraphQLInterfaceType"]] = None,
        resolve_type: Optional[GraphQLTypeResolver] = None,
        description: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[InterfaceTypeDefinitionNode] = None,
        extension_ast_nodes: Optional[Collection[InterfaceTypeExtensionNode]] = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        if resolve_type is not None and not callable(resolve_type):
            raise TypeError(
                f"{name} must provide 'resolve_type' as a function,"
                f" but got: {inspect(resolve_type)}."
            )
        if ast_node and not isinstance(ast_node, InterfaceTypeDefinitionNode):
            raise TypeError(f"{name} AST node must be an InterfaceTypeDefinitionNode.")
        if extension_ast_nodes and not all(
            isinstance(node, InterfaceTypeExtensionNode) for node in extension_ast_nodes
        ):
            raise TypeError(
                f"{name} extension AST nodes must be specified"
                " as a collection of InterfaceTypeExtensionNode instances."
            )
        self._fields = fields
        self._interfaces = interfaces
        self.resolve_type = resolve_type

    def to_kwargs(self) -> GraphQLInterfaceTypeKwargs:
        # noinspection PyArgumentList
        return GraphQLInterfaceTypeKwargs(  # type: ignore
            super().to_kwargs(),
            fields=self.fields.copy(),
            interfaces=self.interfaces,
            resolve_type=self.resolve_type,
        )

    def __copy__(self) -> "GraphQLInterfaceType":  # pragma: no cover
        return self.__class__(**self.to_kwargs())

    @cached_property
    def fields(self) -> GraphQLFieldMap:
        """Get provided fields, wrapping them as GraphQLFields if needed."""
        try:
            fields = resolve_thunk(self._fields)
        except Exception as error:
            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
            raise cls(f"{self.name} fields cannot be resolved. {error}") from error
        if not isinstance(fields, Mapping) or not all(
            isinstance(key, str) for key in fields
        ):
            raise TypeError(
                f"{self.name} fields must be specified"
                " as a mapping with field names as keys."
            )
        if not all(
            isinstance(value, GraphQLField) or is_output_type(value)
            for value in fields.values()
        ):
            raise TypeError(
                f"{self.name} fields must be GraphQLField or output type objects."
            )
        return {
            assert_name(name): value
            if isinstance(value, GraphQLField)
            else GraphQLField(value)  # type: ignore
            for name, value in fields.items()
        }

    @cached_property
    def interfaces(self) -> Tuple["GraphQLInterfaceType", ...]:
        """Get provided interfaces."""
        try:
            interfaces: Collection["GraphQLInterfaceType"] = resolve_thunk(
                self._interfaces  # type: ignore
            )
        except Exception as error:
            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
            raise cls(f"{self.name} interfaces cannot be resolved. {error}") from error
        if interfaces is None:
            interfaces = ()
        elif not is_collection(interfaces) or not all(
            isinstance(value, GraphQLInterfaceType) for value in interfaces
        ):
            raise TypeError(
                f"{self.name} interfaces must be specified"
                " as a collection of GraphQLInterfaceType instances."
            )
        return tuple(interfaces)


def is_interface_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLInterfaceType)


def assert_interface_type(type_: Any) -> GraphQLInterfaceType:
    if not is_interface_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL Interface type.")
    return cast(GraphQLInterfaceType, type_)


class GraphQLUnionTypeKwargs(GraphQLNamedTypeKwargs, total=False):
    types: Tuple[GraphQLObjectType, ...]
    resolve_type: Optional[GraphQLTypeResolver]


class GraphQLUnionType(GraphQLNamedType):
    """Union Type Definition

    When a field can return one of a heterogeneous set of types, a Union type is used
    to describe what types are possible as well as providing a function to determine
    which type is actually used when the field is resolved.

    Example::

        def resolve_type(obj, _info, _type):
            if isinstance(obj, Dog):
                return DogType()
            if isinstance(obj, Cat):
                return CatType()

        PetType = GraphQLUnionType('Pet', [DogType, CatType], resolve_type)
    """

    resolve_type: Optional[GraphQLTypeResolver]
    ast_node: Optional[UnionTypeDefinitionNode]
    extension_ast_nodes: Tuple[UnionTypeExtensionNode, ...]

    def __init__(
        self,
        name: str,
        types: ThunkCollection[GraphQLObjectType],
        resolve_type: Optional[GraphQLTypeResolver] = None,
        description: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[UnionTypeDefinitionNode] = None,
        extension_ast_nodes: Optional[Collection[UnionTypeExtensionNode]] = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        if resolve_type is not None and not callable(resolve_type):
            raise TypeError(
                f"{name} must provide 'resolve_type' as a function,"
                f" but got: {inspect(resolve_type)}."
            )
        if ast_node and not isinstance(ast_node, UnionTypeDefinitionNode):
            raise TypeError(f"{name} AST node must be a UnionTypeDefinitionNode.")
        if extension_ast_nodes and not all(
            isinstance(node, UnionTypeExtensionNode) for node in extension_ast_nodes
        ):
            raise TypeError(
                f"{name} extension AST nodes must be specified"
                " as a collection of UnionTypeExtensionNode instances."
            )
        self._types = types
        self.resolve_type = resolve_type

    def to_kwargs(self) -> GraphQLUnionTypeKwargs:
        # noinspection PyArgumentList
        return GraphQLUnionTypeKwargs(  # type: ignore
            super().to_kwargs(), types=self.types, resolve_type=self.resolve_type
        )

    def __copy__(self) -> "GraphQLUnionType":  # pragma: no cover
        return self.__class__(**self.to_kwargs())

    @cached_property
    def types(self) -> Tuple[GraphQLObjectType, ...]:
        """Get provided types."""
        try:
            types: Collection[GraphQLObjectType] = resolve_thunk(self._types)
        except Exception as error:
            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
            raise cls(f"{self.name} types cannot be resolved. {error}") from error
        if types is None:
            types = ()
        elif not is_collection(types) or not all(
            isinstance(value, GraphQLObjectType) for value in types
        ):
            raise TypeError(
                f"{self.name} types must be specified"
                " as a collection of GraphQLObjectType instances."
            )
        return tuple(types)


def is_union_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLUnionType)


def assert_union_type(type_: Any) -> GraphQLUnionType:
    if not is_union_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL Union type.")
    return cast(GraphQLUnionType, type_)


GraphQLEnumValueMap = Dict[str, "GraphQLEnumValue"]


class GraphQLEnumTypeKwargs(GraphQLNamedTypeKwargs, total=False):
    values: GraphQLEnumValueMap
    names_as_values: Optional[bool]


class GraphQLEnumType(GraphQLNamedType):
    """Enum Type Definition

    Some leaf values of requests and input values are Enums. GraphQL serializes Enum
    values as strings, however internally Enums can be represented by any kind of type,
    often integers. They can also be provided as a Python Enum. In this case, the flag
    `names_as_values` determines what will be used as internal representation. The
    default value of `False` will use the enum values, the value `True` will use the
    enum names, and the value `None` will use the members themselves.

    Example::

        RGBType = GraphQLEnumType('RGB', {
            'RED': 0,
            'GREEN': 1,
            'BLUE': 2
        })

    Example using a Python Enum::

        class RGBEnum(enum.Enum):
            RED = 0
            GREEN = 1
            BLUE = 2

        RGBType = GraphQLEnumType('RGB', enum.Enum)

    Instead of raw values, you can also specify GraphQLEnumValue objects with more
    detail like description or deprecation information.

    Note: If a value is not provided in a definition, the name of the enum value will
    be used as its internal value when the value is serialized.
    """

    values: GraphQLEnumValueMap
    ast_node: Optional[EnumTypeDefinitionNode]
    extension_ast_nodes: Tuple[EnumTypeExtensionNode, ...]

    def __init__(
        self,
        name: str,
        values: Union[GraphQLEnumValueMap, Mapping[str, Any], Type[Enum]],
        names_as_values: Optional[bool] = False,
        description: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[EnumTypeDefinitionNode] = None,
        extension_ast_nodes: Optional[Collection[EnumTypeExtensionNode]] = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        try:  # check for enum
            values = cast(Enum, values).__members__  # type: ignore
        except AttributeError:
            if not isinstance(values, Mapping) or not all(
                isinstance(name, str) for name in values
            ):
                try:
                    # noinspection PyTypeChecker
                    values = dict(values)  # type: ignore
                except (TypeError, ValueError):
                    raise TypeError(
                        f"{name} values must be an Enum or a mapping"
                        " with value names as keys."
                    )
            values = cast(Dict[str, Any], values)
        else:
            values = cast(Dict[str, Enum], values)
            if names_as_values is False:
                values = {key: value.value for key, value in values.items()}
            elif names_as_values is True:
                values = {key: key for key in values}
        values = {
            assert_enum_value_name(key): value
            if isinstance(value, GraphQLEnumValue)
            else GraphQLEnumValue(value)
            for key, value in values.items()
        }
        if ast_node and not isinstance(ast_node, EnumTypeDefinitionNode):
            raise TypeError(f"{name} AST node must be an EnumTypeDefinitionNode.")
        if extension_ast_nodes and not all(
            isinstance(node, EnumTypeExtensionNode) for node in extension_ast_nodes
        ):
            raise TypeError(
                f"{name} extension AST nodes must be specified"
                " as a collection of EnumTypeExtensionNode instances."
            )
        self.values = values

    def to_kwargs(self) -> GraphQLEnumTypeKwargs:
        # noinspection PyArgumentList
        return GraphQLEnumTypeKwargs(  # type: ignore
            super().to_kwargs(), values=self.values.copy()
        )

    def __copy__(self) -> "GraphQLEnumType":  # pragma: no cover
        return self.__class__(**self.to_kwargs())

    @cached_property
    def _value_lookup(self) -> Dict[Any, str]:
        # use first value or name as lookup
        lookup: Dict[Any, str] = {}
        for name, enum_value in self.values.items():
            value = enum_value.value
            if value is None or value is Undefined:
                value = name
            try:
                if value not in lookup:
                    lookup[value] = name
            except TypeError:
                pass  # ignore unhashable values
        return lookup

    def serialize(self, output_value: Any) -> str:
        try:
            return self._value_lookup[output_value]
        except KeyError:  # hashable value not found
            pass
        except TypeError:  # unhashable value, we need to scan all values
            for enum_name, enum_value in self.values.items():
                if enum_value.value == output_value:
                    return enum_name
        raise GraphQLError(
            f"Enum '{self.name}' cannot represent value: {inspect(output_value)}"
        )

    def parse_value(self, input_value: str) -> Any:
        if isinstance(input_value, str):
            try:
                enum_value = self.values[input_value]
            except KeyError:
                raise GraphQLError(
                    f"Value '{input_value}' does not exist in '{self.name}' enum."
                    + did_you_mean_enum_value(self, input_value)
                )
            return enum_value.value
        value_str = inspect(input_value)
        raise GraphQLError(
            f"Enum '{self.name}' cannot represent non-string value: {value_str}."
            + did_you_mean_enum_value(self, value_str)
        )

    def parse_literal(
        self, value_node: ValueNode, _variables: Optional[Dict[str, Any]] = None
    ) -> Any:
        # Note: variables will be resolved before calling this method.
        if isinstance(value_node, EnumValueNode):
            try:
                enum_value = self.values[value_node.value]
            except KeyError:
                value_str = print_ast(value_node)
                raise GraphQLError(
                    f"Value '{value_str}' does not exist in '{self.name}' enum."
                    + did_you_mean_enum_value(self, value_str),
                    value_node,
                )
            return enum_value.value
        value_str = print_ast(value_node)
        raise GraphQLError(
            f"Enum '{self.name}' cannot represent non-enum value: {value_str}."
            + did_you_mean_enum_value(self, value_str),
            value_node,
        )


def is_enum_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLEnumType)


def assert_enum_type(type_: Any) -> GraphQLEnumType:
    if not is_enum_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL Enum type.")
    return cast(GraphQLEnumType, type_)


def did_you_mean_enum_value(enum_type: GraphQLEnumType, unknown_value_str: str) -> str:
    suggested_values = suggestion_list(unknown_value_str, enum_type.values)
    return did_you_mean(suggested_values, "the enum value")


class GraphQLEnumValueKwargs(TypedDict, total=False):
    value: Any
    description: Optional[str]
    deprecation_reason: Optional[str]
    extensions: Dict[str, Any]
    ast_node: Optional[EnumValueDefinitionNode]


class GraphQLEnumValue:

    value: Any
    description: Optional[str]
    deprecation_reason: Optional[str]
    extensions: Dict[str, Any]
    ast_node: Optional[EnumValueDefinitionNode]

    def __init__(
        self,
        value: Any = None,
        description: Optional[str] = None,
        deprecation_reason: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[EnumValueDefinitionNode] = None,
    ) -> None:
        if description is not None and not is_description(description):
            raise TypeError("The description of the enum value must be a string.")
        if deprecation_reason is not None and not is_description(deprecation_reason):
            raise TypeError(
                "The deprecation reason for the enum value must be a string."
            )
        if extensions is None:
            extensions = {}
        elif not isinstance(extensions, dict) or not all(
            isinstance(key, str) for key in extensions
        ):
            raise TypeError(
                "Enum value extensions must be a dictionary with string keys."
            )
        if ast_node and not isinstance(ast_node, EnumValueDefinitionNode):
            raise TypeError("AST node must be an EnumValueDefinitionNode.")
        self.value = value
        self.description = description
        self.deprecation_reason = deprecation_reason
        self.extensions = extensions
        self.ast_node = ast_node

    def __eq__(self, other: Any) -> bool:
        return self is other or (
            isinstance(other, GraphQLEnumValue)
            and self.value == other.value
            and self.description == other.description
            and self.deprecation_reason == other.deprecation_reason
            and self.extensions == other.extensions
        )

    def to_kwargs(self) -> GraphQLEnumValueKwargs:
        return GraphQLEnumValueKwargs(
            value=self.value,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            extensions=self.extensions,
            ast_node=self.ast_node,
        )

    def __copy__(self) -> "GraphQLEnumValue":  # pragma: no cover
        return self.__class__(**self.to_kwargs())


GraphQLInputFieldMap = Dict[str, "GraphQLInputField"]
GraphQLInputFieldOutType = Callable[[Dict[str, Any]], Any]


class GraphQLInputObjectTypeKwargs(GraphQLNamedTypeKwargs, total=False):
    fields: GraphQLInputFieldMap
    out_type: Optional[GraphQLInputFieldOutType]


class GraphQLInputObjectType(GraphQLNamedType):
    """Input Object Type Definition

    An input object defines a structured collection of fields which may be supplied
    to a field argument.

    Using ``NonNull`` will ensure that a value must be provided by the query.

    Example::

        NonNullFloat = GraphQLNonNull(GraphQLFloat())

        class GeoPoint(GraphQLInputObjectType):
            name = 'GeoPoint'
            fields = {
                'lat': GraphQLInputField(NonNullFloat),
                'lon': GraphQLInputField(NonNullFloat),
                'alt': GraphQLInputField(
                          GraphQLFloat(), default_value=0)
            }

    The outbound values will be Python dictionaries by default, but you can have them
    converted to other types by specifying an ``out_type`` function or class.
    """

    ast_node: Optional[InputObjectTypeDefinitionNode]
    extension_ast_nodes: Tuple[InputObjectTypeExtensionNode, ...]

    def __init__(
        self,
        name: str,
        fields: ThunkMapping["GraphQLInputField"],
        description: Optional[str] = None,
        out_type: Optional[GraphQLInputFieldOutType] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[InputObjectTypeDefinitionNode] = None,
        extension_ast_nodes: Optional[Collection[InputObjectTypeExtensionNode]] = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            extensions=extensions,
            ast_node=ast_node,
            extension_ast_nodes=extension_ast_nodes,
        )
        if out_type is not None and not callable(out_type):
            raise TypeError(f"The out type for {name} must be a function or a class.")
        if ast_node and not isinstance(ast_node, InputObjectTypeDefinitionNode):
            raise TypeError(
                f"{name} AST node must be an InputObjectTypeDefinitionNode."
            )
        if extension_ast_nodes and not all(
            isinstance(node, InputObjectTypeExtensionNode)
            for node in extension_ast_nodes
        ):
            raise TypeError(
                f"{name} extension AST nodes must be specified"
                " as a collection of InputObjectTypeExtensionNode instances."
            )
        self._fields = fields
        if out_type is not None:
            self.out_type = out_type  # type: ignore

    @staticmethod
    def out_type(value: Dict[str, Any]) -> Any:
        """Transform outbound values (this is an extension of GraphQL.js).

        This default implementation passes values unaltered as dictionaries.
        """
        return value

    def to_kwargs(self) -> GraphQLInputObjectTypeKwargs:
        # noinspection PyArgumentList
        return GraphQLInputObjectTypeKwargs(  # type: ignore
            super().to_kwargs(),
            fields=self.fields.copy(),
            out_type=None
            if self.out_type is GraphQLInputObjectType.out_type
            else self.out_type,
        )

    def __copy__(self) -> "GraphQLInputObjectType":  # pragma: no cover
        return self.__class__(**self.to_kwargs())

    @cached_property
    def fields(self) -> GraphQLInputFieldMap:
        """Get provided fields, wrap them as GraphQLInputField if needed."""
        try:
            fields = resolve_thunk(self._fields)
        except Exception as error:
            cls = GraphQLError if isinstance(error, GraphQLError) else TypeError
            raise cls(f"{self.name} fields cannot be resolved. {error}") from error
        if not isinstance(fields, Mapping) or not all(
            isinstance(key, str) for key in fields
        ):
            raise TypeError(
                f"{self.name} fields must be specified"
                " as a mapping with field names as keys."
            )
        if not all(
            isinstance(value, GraphQLInputField) or is_input_type(value)
            for value in fields.values()
        ):
            raise TypeError(
                f"{self.name} fields must be"
                " GraphQLInputField or input type objects."
            )
        return {
            assert_name(name): value
            if isinstance(value, GraphQLInputField)
            else GraphQLInputField(value)  # type: ignore
            for name, value in fields.items()
        }


def is_input_object_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLInputObjectType)


def assert_input_object_type(type_: Any) -> GraphQLInputObjectType:
    if not is_input_object_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL Input Object type.")
    return cast(GraphQLInputObjectType, type_)


class GraphQLInputFieldKwargs(TypedDict, total=False):
    type_: "GraphQLInputType"
    default_value: Any
    description: Optional[str]
    deprecation_reason: Optional[str]
    out_name: Optional[str]
    extensions: Dict[str, Any]
    ast_node: Optional[InputValueDefinitionNode]


class GraphQLInputField:
    """Definition of a GraphQL input field"""

    type: "GraphQLInputType"
    default_value: Any
    description: Optional[str]
    deprecation_reason: Optional[str]
    out_name: Optional[str]  # for transforming names (extension of GraphQL.js)
    extensions: Dict[str, Any]
    ast_node: Optional[InputValueDefinitionNode]

    def __init__(
        self,
        type_: "GraphQLInputType",
        default_value: Any = Undefined,
        description: Optional[str] = None,
        deprecation_reason: Optional[str] = None,
        out_name: Optional[str] = None,
        extensions: Optional[Dict[str, Any]] = None,
        ast_node: Optional[InputValueDefinitionNode] = None,
    ) -> None:
        if not is_input_type(type_):
            raise TypeError("Input field type must be a GraphQL input type.")
        if description is not None and not is_description(description):
            raise TypeError("Input field description must be a string.")
        if deprecation_reason is not None and not is_description(deprecation_reason):
            raise TypeError("Input field deprecation reason must be a string.")
        if out_name is not None and not isinstance(out_name, str):
            raise TypeError("Input field out name must be a string.")
        if extensions is None:
            extensions = {}
        elif not isinstance(extensions, dict) or not all(
            isinstance(key, str) for key in extensions
        ):
            raise TypeError(
                "Input field extensions must be a dictionary with string keys."
            )
        if ast_node and not isinstance(ast_node, InputValueDefinitionNode):
            raise TypeError("Input field AST node must be an InputValueDefinitionNode.")
        self.type = type_
        self.default_value = default_value
        self.description = description
        self.deprecation_reason = deprecation_reason
        self.out_name = out_name
        self.extensions = extensions
        self.ast_node = ast_node

    def __eq__(self, other: Any) -> bool:
        return self is other or (
            isinstance(other, GraphQLInputField)
            and self.type == other.type
            and self.default_value == other.default_value
            and self.description == other.description
            and self.deprecation_reason == other.deprecation_reason
            and self.extensions == other.extensions
            and self.out_name == other.out_name
        )

    def to_kwargs(self) -> GraphQLInputFieldKwargs:
        return GraphQLInputFieldKwargs(
            type_=self.type,
            default_value=self.default_value,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            out_name=self.out_name,
            extensions=self.extensions,
            ast_node=self.ast_node,
        )

    def __copy__(self) -> "GraphQLInputField":  # pragma: no cover
        return self.__class__(**self.to_kwargs())


def is_required_input_field(field: GraphQLInputField) -> bool:
    return is_non_null_type(field.type) and field.default_value is Undefined


# Wrapper types


class GraphQLList(Generic[GT], GraphQLWrappingType[GT]):
    """List Type Wrapper

    A list is a wrapping type which points to another type. Lists are often created
    within the context of defining the fields of an object type.

    Example::

        class PersonType(GraphQLObjectType):
            name = 'Person'

            @property
            def fields(self):
                return {
                    'parents': GraphQLField(GraphQLList(PersonType())),
                    'children': GraphQLField(GraphQLList(PersonType())),
                }
    """

    def __init__(self, type_: GT) -> None:
        super().__init__(type_=type_)

    def __str__(self) -> str:
        return f"[{self.of_type}]"


def is_list_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLList)


def assert_list_type(type_: Any) -> GraphQLList:
    if not is_list_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL List type.")
    return cast(GraphQLList, type_)


GNT = TypeVar("GNT", bound="GraphQLNullableType")


class GraphQLNonNull(GraphQLWrappingType[GNT], Generic[GNT]):
    """Non-Null Type Wrapper

    A non-null is a wrapping type which points to another type. Non-null types enforce
    that their values are never null and can ensure an error is raised if this ever
    occurs during a request. It is useful for fields which you can make a strong
    guarantee on non-nullability, for example usually the id field of a database row
    will never be null.

    Example::

        class RowType(GraphQLObjectType):
            name = 'Row'
            fields = {
                'id': GraphQLField(GraphQLNonNull(GraphQLString()))
            }

    Note: the enforcement of non-nullability occurs within the executor.
    """

    def __init__(self, type_: GNT):
        super().__init__(type_=type_)
        if isinstance(type_, GraphQLNonNull):
            raise TypeError(
                "Can only create NonNull of a Nullable GraphQLType but got:"
                f" {type_}."
            )

    def __str__(self) -> str:
        return f"{self.of_type}!"


def is_non_null_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLNonNull)


def assert_non_null_type(type_: Any) -> GraphQLNonNull:
    if not is_non_null_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL Non-Null type.")
    return cast(GraphQLNonNull, type_)


# These types can all accept null as a value.

graphql_nullable_types = (
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLList,
)

GraphQLNullableType = Union[
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLList,
]


def is_nullable_type(type_: Any) -> bool:
    return isinstance(type_, graphql_nullable_types)


def assert_nullable_type(type_: Any) -> GraphQLNullableType:
    if not is_nullable_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL nullable type.")
    return cast(GraphQLNullableType, type_)


@overload
def get_nullable_type(type_: None) -> None:
    ...


@overload
def get_nullable_type(type_: GraphQLNullableType) -> GraphQLNullableType:
    ...


@overload
def get_nullable_type(type_: GraphQLNonNull) -> GraphQLNullableType:
    ...


def get_nullable_type(
    type_: Optional[Union[GraphQLNullableType, GraphQLNonNull]]
) -> Optional[GraphQLNullableType]:
    """Unwrap possible non-null type"""
    if is_non_null_type(type_):
        type_ = cast(GraphQLNonNull, type_)
        type_ = type_.of_type
    return cast(Optional[GraphQLNullableType], type_)


# These types may be used as input types for arguments and directives.

graphql_input_types = (GraphQLScalarType, GraphQLEnumType, GraphQLInputObjectType)

GraphQLInputType = Union[
    GraphQLScalarType, GraphQLEnumType, GraphQLInputObjectType, GraphQLWrappingType
]


def is_input_type(type_: Any) -> bool:
    return isinstance(type_, graphql_input_types) or (
        isinstance(type_, GraphQLWrappingType) and is_input_type(type_.of_type)
    )


def assert_input_type(type_: Any) -> GraphQLInputType:
    if not is_input_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL input type.")
    return cast(GraphQLInputType, type_)


# These types may be used as output types as the result of fields.

graphql_output_types = (
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
)

GraphQLOutputType = Union[
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
    GraphQLWrappingType,
]


def is_output_type(type_: Any) -> bool:
    return isinstance(type_, graphql_output_types) or (
        isinstance(type_, GraphQLWrappingType) and is_output_type(type_.of_type)
    )


def assert_output_type(type_: Any) -> GraphQLOutputType:
    if not is_output_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL output type.")
    return cast(GraphQLOutputType, type_)


# These named types do not include modifiers like List or NonNull.

GraphQLNamedInputType = Union[
    GraphQLScalarType, GraphQLEnumType, GraphQLInputObjectType
]

GraphQLNamedOutputType = Union[
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
]


def is_named_type(type_: Any) -> bool:
    return isinstance(type_, GraphQLNamedType)


def assert_named_type(type_: Any) -> GraphQLNamedType:
    if not is_named_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL named type.")
    return cast(GraphQLNamedType, type_)


@overload
def get_named_type(type_: None) -> None:
    ...


@overload
def get_named_type(type_: GraphQLType) -> GraphQLNamedType:
    ...


def get_named_type(type_: Optional[GraphQLType]) -> Optional[GraphQLNamedType]:
    """Unwrap possible wrapping type"""
    if type_:
        unwrapped_type = type_
        while is_wrapping_type(unwrapped_type):
            unwrapped_type = cast(GraphQLWrappingType, unwrapped_type)
            unwrapped_type = unwrapped_type.of_type
        return cast(GraphQLNamedType, unwrapped_type)
    return None


# These types may describe types which may be leaf values.

graphql_leaf_types = (GraphQLScalarType, GraphQLEnumType)

GraphQLLeafType = Union[GraphQLScalarType, GraphQLEnumType]


def is_leaf_type(type_: Any) -> bool:
    return isinstance(type_, graphql_leaf_types)


def assert_leaf_type(type_: Any) -> GraphQLLeafType:
    if not is_leaf_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL leaf type.")
    return cast(GraphQLLeafType, type_)


# These types may describe the parent context of a selection set.

graphql_composite_types = (GraphQLObjectType, GraphQLInterfaceType, GraphQLUnionType)

GraphQLCompositeType = Union[GraphQLObjectType, GraphQLInterfaceType, GraphQLUnionType]


def is_composite_type(type_: Any) -> bool:
    return isinstance(type_, graphql_composite_types)


def assert_composite_type(type_: Any) -> GraphQLType:
    if not is_composite_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL composite type.")
    return cast(GraphQLType, type_)


# These types may describe abstract types.

graphql_abstract_types = (GraphQLInterfaceType, GraphQLUnionType)

GraphQLAbstractType = Union[GraphQLInterfaceType, GraphQLUnionType]


def is_abstract_type(type_: Any) -> bool:
    return isinstance(type_, graphql_abstract_types)


def assert_abstract_type(type_: Any) -> GraphQLAbstractType:
    if not is_abstract_type(type_):
        raise TypeError(f"Expected {type_} to be a GraphQL composite type.")
    return cast(GraphQLAbstractType, type_)



from copy import copy, deepcopy

from pytest import raises

from graphql.pyutils import FrozenError, FrozenList


def describe_frozen_list():
    def can_read():
        fl = FrozenList([1, 2, 3])
        assert fl == [1, 2, 3]
        assert list(i for i in fl) == fl
        assert fl.copy() == fl
        assert 2 in fl
        assert 4 not in fl
        assert fl + [4, 5] == [1, 2, 3, 4, 5]
        assert [4, 5] + fl == [4, 5, 1, 2, 3]
        assert fl * 2 == [1, 2, 3, 1, 2, 3]
        assert 2 * fl == [1, 2, 3, 1, 2, 3]
        assert fl[1] == 2
        with raises(IndexError):
            fl[3]
        assert fl[1:4] == [2, 3]
        assert fl[::2] == [1, 3]
        assert len(fl) == 3
        assert min(fl) == 1
        assert max(fl) == 3
        assert sum(fl) == 6
        assert fl.index(2) == 1
        with raises(ValueError):
            fl.index(4)
        assert fl.count(2) == 1
        assert fl.count(4) == 0
        assert list(reversed(fl)) == [3, 2, 1]
        assert sorted(fl) == [1, 2, 3]

    def cannot_write():
        fl = FrozenList([1, 2, 3])
        with raises(FrozenError):
            fl[1] = 4
        with raises(FrozenError):
            fl[1:4] = [4]
        with raises(FrozenError):
            del fl[1]
        with raises(FrozenError):
            del fl[1:4]
        with raises(FrozenError):
            fl[1::2] = [4]
        with raises(FrozenError):
            del fl[::2]
        with raises(FrozenError):
            fl.append(4)
        with raises(FrozenError):
            fl.clear()
        with raises(FrozenError):
            fl.extend([4])
        with raises(FrozenError):
            fl += [4]
        with raises(FrozenError):
            fl *= 2
        with raises(FrozenError):
            fl.insert(1, 4)
        with raises(FrozenError):
            fl.pop()
        with raises(FrozenError):
            fl.remove(2)
        with raises(FrozenError):
            fl.sort()
        with raises(FrozenError):
            fl.reverse()
        assert fl == [1, 2, 3]

    def can_add_rol():
        fl1 = FrozenList([1, 2])
        rol2 = FrozenList([3, 4])
        assert fl1 + rol2 == [1, 2, 3, 4]

    def can_add_tuple():
        fl = FrozenList([1, 2])
        assert fl + (3, 4) == [1, 2, 3, 4]

    def can_hash():
        fl1 = FrozenList([1, 2])
        fl2 = FrozenList([1, 2])
        assert fl2 == fl1
        assert fl2 is not fl1
        assert hash(fl2) == hash(fl1)
        fl3 = FrozenList([1, 3])
        assert fl3 != fl1
        assert hash(fl3) != hash(fl1)

    def can_copy():
        fl1 = FrozenList([1, 2])
        fl2 = copy(fl1)
        assert isinstance(fl2, FrozenList)
        assert fl2 == fl1
        assert hash(fl2) == hash(fl1)
        assert fl2 is not fl1

    def can_deep_copy():
        fl11 = FrozenList([1, 2])
        fl12 = FrozenList([2, 1])
        fl1 = FrozenList([fl11, fl12])
        fl2 = deepcopy(fl1)
        assert isinstance(fl2, FrozenList)
        assert fl2 == fl1
        assert hash(fl2) == hash(fl1)
        assert isinstance(fl2[0], FrozenList)
        assert isinstance(fl2[1], FrozenList)
        assert fl2[0] == fl1[0]
        assert fl2[0] is not fl1[0]
        assert fl2[1] == fl1[1]
        assert fl2[1] is not fl1[1]



from graphql.pyutils import camel_to_snake, snake_to_camel


def describe_camel_to_snake():
    def converts_typical_names():
        assert camel_to_snake("CamelCase") == "camel_case"
        assert (
            camel_to_snake("InputObjectTypeExtensionNode")
            == "input_object_type_extension_node"
        )
        assert camel_to_snake("CamelToSnake") == "camel_to_snake"

    def may_start_with_lowercase():
        assert camel_to_snake("camelCase") == "camel_case"

    def works_with_acronyms():
        assert camel_to_snake("SlowXMLParser") == "slow_xml_parser"
        assert camel_to_snake("FastGraphQLParser") == "fast_graph_ql_parser"

    def works_with_numbers():
        assert camel_to_snake("Python3Script") == "python3_script"
        assert camel_to_snake("camel2snake") == "camel2snake"

    def keeps_already_snake():
        assert camel_to_snake("snake_case") == "snake_case"


def describe_snake_to_camel():
    def converts_typical_names():
        assert snake_to_camel("snake_case") == "SnakeCase"
        assert (
            snake_to_camel("input_object_type_extension_node")
            == "InputObjectTypeExtensionNode"
        )
        assert snake_to_camel("snake_to_camel") == "SnakeToCamel"

    def may_start_with_uppercase():
        assert snake_to_camel("Snake_case") == "SnakeCase"

    def works_with_acronyms():
        assert snake_to_camel("slow_xml_parser") == "SlowXmlParser"
        assert snake_to_camel("fast_graph_ql_parser") == "FastGraphQlParser"

    def works_with_numbers():
        assert snake_to_camel("python3_script") == "Python3Script"
        assert snake_to_camel("snake2camel") == "Snake2camel"

    def keeps_already_camel():
        assert snake_to_camel("CamelCase") == "CamelCase"

    def can_produce_lower_camel_case():
        assert snake_to_camel("snake_case", upper=False) == "snakeCase"
        assert (
            snake_to_camel("input_object_type_extension_node", False)
            == "inputObjectTypeExtensionNode"
        )



from copy import deepcopy

from pytest import raises

from graphql.language import ScalarTypeDefinitionNode, NameNode, print_ast, parse

from ..fixtures import kitchen_sink_sdl  # noqa: F401
from ..utils import dedent


def describe_printer_sdl_document():
    def prints_minimal_ast():
        node = ScalarTypeDefinitionNode(name=NameNode(value="foo"))
        assert print_ast(node) == "scalar foo"

    def produces_helpful_error_messages():
        bad_ast = {"random": "Data"}
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            print_ast(bad_ast)  # type: ignore
        msg = str(exc_info.value)
        assert msg == "Not an AST Node: {'random': 'Data'}."

    # noinspection PyShadowingNames
    def prints_kitchen_sink_without_altering_ast(kitchen_sink_sdl):  # noqa: F811
        ast = parse(kitchen_sink_sdl, no_location=True)

        ast_before_print_call = deepcopy(ast)
        printed = print_ast(ast)
        printed_ast = parse(printed, no_location=True)

        assert printed_ast == ast
        assert deepcopy(ast) == ast_before_print_call

        assert printed == dedent(
            '''
            """This is a description of the schema as a whole."""
            schema {
              query: QueryType
              mutation: MutationType
            }

            """
            This is a description
            of the `Foo` type.
            """
            type Foo implements Bar & Baz & Two {
              "Description of the `one` field."
              one: Type
              """This is a description of the `two` field."""
              two(
                """This is a description of the `argument` argument."""
                argument: InputType!
              ): Type
              """This is a description of the `three` field."""
              three(argument: InputType, other: String): Int
              four(argument: String = "string"): String
              five(argument: [String] = ["string", "string"]): String
              six(argument: InputType = {key: "value"}): Type
              seven(argument: Int = null): Type
            }

            type AnnotatedObject @onObject(arg: "value") {
              annotatedField(arg: Type = "default" @onArgumentDefinition): Type @onField
            }

            type UndefinedType

            extend type Foo {
              seven(argument: [String]): Type
            }

            extend type Foo @onType

            interface Bar {
              one: Type
              four(argument: String = "string"): String
            }

            interface AnnotatedInterface @onInterface {
              annotatedField(arg: Type @onArgumentDefinition): Type @onField
            }

            interface UndefinedInterface

            extend interface Bar implements Two {
              two(argument: InputType!): Type
            }

            extend interface Bar @onInterface

            interface Baz implements Bar & Two {
              one: Type
              two(argument: InputType!): Type
              four(argument: String = "string"): String
            }

            union Feed = Story | Article | Advert

            union AnnotatedUnion @onUnion = A | B

            union AnnotatedUnionTwo @onUnion = A | B

            union UndefinedUnion

            extend union Feed = Photo | Video

            extend union Feed @onUnion

            scalar CustomScalar

            scalar AnnotatedScalar @onScalar

            extend scalar CustomScalar @onScalar

            enum Site {
              """This is a description of the `DESKTOP` value"""
              DESKTOP
              """This is a description of the `MOBILE` value"""
              MOBILE
              "This is a description of the `WEB` value"
              WEB
            }

            enum AnnotatedEnum @onEnum {
              ANNOTATED_VALUE @onEnumValue
              OTHER_VALUE
            }

            enum UndefinedEnum

            extend enum Site {
              VR
            }

            extend enum Site @onEnum

            input InputType {
              key: String!
              answer: Int = 42
            }

            input AnnotatedInput @onInputObject {
              annotatedField: Type @onInputFieldDefinition
            }

            input UndefinedInput

            extend input InputType {
              other: Float = 1.23e4 @onInputFieldDefinition
            }

            extend input InputType @onInputObject

            """This is a description of the `@skip` directive"""
            directive @skip(
              """This is a description of the `if` argument"""
              if: Boolean! @onArgumentDefinition
            ) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

            directive @include(if: Boolean!) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

            directive @include2(if: Boolean!) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

            directive @myRepeatableDir(name: String!) repeatable on OBJECT | INTERFACE

            extend schema @onSchema

            extend schema @onSchema {
              subscription: SubscriptionType
            }
            '''  # noqa: E501
        )



from textwrap import dedent
from typing import Any, Dict, List, Optional, Union

from ..language import DirectiveLocation

try:
    from typing import TypedDict, Literal
except ImportError:  # Python < 3.8
    from typing_extensions import TypedDict, Literal  # type: ignore

__all__ = [
    "get_introspection_query",
    "IntrospectionDirective",
    "IntrospectionEnumType",
    "IntrospectionField",
    "IntrospectionInputObjectType",
    "IntrospectionInputValue",
    "IntrospectionInterfaceType",
    "IntrospectionListType",
    "IntrospectionNonNullType",
    "IntrospectionObjectType",
    "IntrospectionQuery",
    "IntrospectionScalarType",
    "IntrospectionSchema",
    "IntrospectionType",
    "IntrospectionTypeRef",
    "IntrospectionUnionType",
]


def get_introspection_query(
    descriptions: bool = True,
    specified_by_url: bool = False,
    directive_is_repeatable: bool = False,
    schema_description: bool = False,
    input_value_deprecation: bool = False,
) -> str:
    """Get a query for introspection.

    Optionally, you can exclude descriptions, include specification URLs,
    include repeatability of directives, and specify whether to include
    the schema description as well.
    """
    maybe_description = "description" if descriptions else ""
    maybe_specified_by_url = "specifiedByURL" if specified_by_url else ""
    maybe_directive_is_repeatable = "isRepeatable" if directive_is_repeatable else ""
    maybe_schema_description = maybe_description if schema_description else ""

    def input_deprecation(string: str) -> Optional[str]:
        return string if input_value_deprecation else ""

    return dedent(
        f"""
        query IntrospectionQuery {{
          __schema {{
            {maybe_schema_description}
            queryType {{ name }}
            mutationType {{ name }}
            subscriptionType {{ name }}
            types {{
              ...FullType
            }}
            directives {{
              name
              {maybe_description}
              {maybe_directive_is_repeatable}
              locations
              args{input_deprecation("(includeDeprecated: true)")} {{
                ...InputValue
              }}
            }}
          }}
        }}

        fragment FullType on __Type {{
          kind
          name
          {maybe_description}
          {maybe_specified_by_url}
          fields(includeDeprecated: true) {{
            name
            {maybe_description}
            args{input_deprecation("(includeDeprecated: true)")} {{
              ...InputValue
            }}
            type {{
              ...TypeRef
            }}
            isDeprecated
            deprecationReason
          }}
          inputFields{input_deprecation("(includeDeprecated: true)")} {{
            ...InputValue
          }}
          interfaces {{
            ...TypeRef
          }}
          enumValues(includeDeprecated: true) {{
            name
            {maybe_description}
            isDeprecated
            deprecationReason
          }}
          possibleTypes {{
            ...TypeRef
          }}
        }}

        fragment InputValue on __InputValue {{
          name
          {maybe_description}
          type {{ ...TypeRef }}
          defaultValue
          {input_deprecation("isDeprecated")}
          {input_deprecation("deprecationReason")}
        }}

        fragment TypeRef on __Type {{
          kind
          name
          ofType {{
            kind
            name
            ofType {{
              kind
              name
              ofType {{
                kind
                name
                ofType {{
                  kind
                  name
                  ofType {{
                    kind
                    name
                    ofType {{
                      kind
                      name
                      ofType {{
                        kind
                        name
                      }}
                    }}
                  }}
                }}
              }}
            }}
          }}
        }}
        """
    )


# Unfortunately, the following type definitions are a bit simplistic
# because of current restrictions in the typing system (mypy):
# - no recursion, see https://github.com/python/mypy/issues/731
# - no generic typed dicts, see https://github.com/python/mypy/issues/3863

# simplified IntrospectionNamedType to avoids cycles
SimpleIntrospectionType = Dict[str, Any]


class MaybeWithDescription(TypedDict, total=False):
    description: Optional[str]


class WithName(MaybeWithDescription):
    name: str


class MaybeWithSpecifiedByUrl(TypedDict, total=False):
    specifiedByURL: Optional[str]


class WithDeprecated(TypedDict):
    isDeprecated: bool
    deprecationReason: Optional[str]


class MaybeWithDeprecated(TypedDict, total=False):
    isDeprecated: bool
    deprecationReason: Optional[str]


class IntrospectionInputValue(WithName, MaybeWithDeprecated):
    type: SimpleIntrospectionType  # should be IntrospectionInputType
    defaultValue: Optional[str]


class IntrospectionField(WithName, WithDeprecated):
    args: List[IntrospectionInputValue]
    type: SimpleIntrospectionType  # should be IntrospectionOutputType


class IntrospectionEnumValue(WithName, WithDeprecated):
    pass


class MaybeWithIsRepeatable(TypedDict, total=False):
    isRepeatable: bool


class IntrospectionDirective(WithName, MaybeWithIsRepeatable):
    locations: List[DirectiveLocation]
    args: List[IntrospectionInputValue]


class IntrospectionScalarType(WithName, MaybeWithSpecifiedByUrl):
    kind: Literal["scalar"]


class IntrospectionInterfaceType(WithName):
    kind: Literal["interface"]
    fields: List[IntrospectionField]
    interfaces: List[SimpleIntrospectionType]  # should be InterfaceType
    possibleTypes: List[SimpleIntrospectionType]  # should be NamedType


class IntrospectionObjectType(WithName):
    kind: Literal["object"]
    fields: List[IntrospectionField]
    interfaces: List[SimpleIntrospectionType]  # should be InterfaceType


class IntrospectionUnionType(WithName):
    kind: Literal["union"]
    possibleTypes: List[SimpleIntrospectionType]  # should be NamedType


class IntrospectionEnumType(WithName):
    kind: Literal["enum"]
    enumValues: List[IntrospectionEnumValue]


class IntrospectionInputObjectType(WithName):
    kind: Literal["input_object"]
    inputFields: List[IntrospectionInputValue]


IntrospectionType = Union[
    IntrospectionScalarType,
    IntrospectionObjectType,
    IntrospectionInterfaceType,
    IntrospectionUnionType,
    IntrospectionEnumType,
    IntrospectionInputObjectType,
]


IntrospectionOutputType = Union[
    IntrospectionScalarType,
    IntrospectionObjectType,
    IntrospectionInterfaceType,
    IntrospectionUnionType,
    IntrospectionEnumType,
]


IntrospectionInputType = Union[
    IntrospectionScalarType, IntrospectionEnumType, IntrospectionInputObjectType
]


class IntrospectionListType(TypedDict):
    kind: Literal["list"]
    ofType: SimpleIntrospectionType  # should be IntrospectionType


class IntrospectionNonNullType(TypedDict):
    kind: Literal["non_null"]
    ofType: SimpleIntrospectionType  # should be IntrospectionType


IntrospectionTypeRef = Union[
    IntrospectionType, IntrospectionListType, IntrospectionNonNullType
]


class IntrospectionSchema(MaybeWithDescription):
    queryType: IntrospectionObjectType
    mutationType: Optional[IntrospectionObjectType]
    subscriptionType: Optional[IntrospectionObjectType]
    types: List[IntrospectionType]
    directives: List[IntrospectionDirective]


class IntrospectionQuery(TypedDict):
    """The root typed dictionary for schema introspections."""

    __schema: IntrospectionSchema



"""Fixtures for graphql tests"""
import json
from os.path import dirname, join

from pytest import fixture

__all__ = [
    "kitchen_sink_query",
    "kitchen_sink_sdl",
    "big_schema_sdl",
    "big_schema_introspection_result",
]


def read_graphql(name):
    path = join(dirname(__file__), name + ".graphql")
    return open(path, encoding="utf-8").read()


def read_json(name):
    path = join(dirname(__file__), name + ".json")
    return json.load(open(path, encoding="utf-8"))


@fixture(scope="module")
def kitchen_sink_query():
    return read_graphql("kitchen_sink")


@fixture(scope="module")
def kitchen_sink_sdl():
    return read_graphql("schema_kitchen_sink")


@fixture(scope="module")
def big_schema_sdl():
    return read_graphql("github_schema")


@fixture(scope="module")
def big_schema_introspection_result():
    return read_json("github_schema")



from math import inf, nan

from pytest import raises

from graphql.error import GraphQLError
from graphql.language import (
    BooleanValueNode,
    EnumValueNode,
    FloatValueNode,
    IntValueNode,
    ListValueNode,
    NameNode,
    NullValueNode,
    ObjectFieldNode,
    ObjectValueNode,
    StringValueNode,
)
from graphql.pyutils import Undefined
from graphql.type import (
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLFloat,
    GraphQLID,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
    GraphQLString,
)
from graphql.utilities import ast_from_value


def describe_ast_from_value():
    def converts_boolean_values_to_asts():
        assert ast_from_value(True, GraphQLBoolean) == BooleanValueNode(value=True)

        assert ast_from_value(False, GraphQLBoolean) == BooleanValueNode(value=False)

        assert ast_from_value(Undefined, GraphQLBoolean) is None

        assert ast_from_value(None, GraphQLBoolean) == NullValueNode()

        assert ast_from_value(0, GraphQLBoolean) == BooleanValueNode(value=False)

        assert ast_from_value(1, GraphQLBoolean) == BooleanValueNode(value=True)

        non_null_boolean = GraphQLNonNull(GraphQLBoolean)
        assert ast_from_value(0, non_null_boolean) == BooleanValueNode(value=False)

    def converts_int_values_to_int_asts():
        assert ast_from_value(-1, GraphQLInt) == IntValueNode(value="-1")

        assert ast_from_value(123.0, GraphQLInt) == IntValueNode(value="123")

        assert ast_from_value(1e4, GraphQLInt) == IntValueNode(value="10000")

        # GraphQL spec does not allow coercing non-integer values to Int to
        # avoid accidental data loss.
        with raises(GraphQLError) as exc_info:
            assert ast_from_value(123.5, GraphQLInt)
        msg = str(exc_info.value)
        assert msg == "Int cannot represent non-integer value: 123.5"

        # Note: outside the bounds of 32bit signed int.
        with raises(GraphQLError) as exc_info:
            assert ast_from_value(1e40, GraphQLInt)
        msg = str(exc_info.value)
        assert msg == "Int cannot represent non 32-bit signed integer value: 1e+40"

        with raises(GraphQLError) as exc_info:
            ast_from_value(nan, GraphQLInt)
        msg = str(exc_info.value)
        assert msg == "Int cannot represent non-integer value: nan"

    def converts_float_values_to_float_asts():
        # luckily in Python we can discern between float and int
        assert ast_from_value(-1, GraphQLFloat) == FloatValueNode(value="-1")

        assert ast_from_value(123.0, GraphQLFloat) == FloatValueNode(value="123")

        assert ast_from_value(123.5, GraphQLFloat) == FloatValueNode(value="123.5")

        assert ast_from_value(1e4, GraphQLFloat) == FloatValueNode(value="10000")

        assert ast_from_value(1e40, GraphQLFloat) == FloatValueNode(value="1e+40")

        assert ast_from_value(123456789.01234567, GraphQLFloat) == FloatValueNode(
            value="123456789.01234567"
        )

        assert ast_from_value(1.1, GraphQLFloat) == FloatValueNode(value="1.1")

    def converts_string_values_to_string_asts():
        assert ast_from_value("hello", GraphQLString) == StringValueNode(value="hello")

        assert ast_from_value("VALUE", GraphQLString) == StringValueNode(value="VALUE")

        assert ast_from_value("VA\nLUE", GraphQLString) == StringValueNode(
            value="VA\nLUE"
        )

        assert ast_from_value(123, GraphQLString) == StringValueNode(value="123")

        assert ast_from_value(False, GraphQLString) == StringValueNode(value="false")

        assert ast_from_value(None, GraphQLString) == NullValueNode()

        assert ast_from_value(Undefined, GraphQLString) is None

    def converts_id_values_to_int_or_string_asts():
        assert ast_from_value("hello", GraphQLID) == StringValueNode(value="hello")

        assert ast_from_value("VALUE", GraphQLID) == StringValueNode(value="VALUE")

        # Note: EnumValues cannot contain non-identifier characters
        assert ast_from_value("VA\nLUE", GraphQLID) == StringValueNode(value="VA\nLUE")

        # Note: IntValues are used when possible.
        assert ast_from_value(-1, GraphQLID) == IntValueNode(value="-1")

        assert ast_from_value(123, GraphQLID) == IntValueNode(value="123")

        assert ast_from_value("123", GraphQLID) == IntValueNode(value="123")

        assert ast_from_value("01", GraphQLID) == StringValueNode(value="01")

        with raises(GraphQLError) as exc_info:
            assert ast_from_value(False, GraphQLID)
        assert str(exc_info.value) == "ID cannot represent value: False"

        assert ast_from_value(None, GraphQLID) == NullValueNode()

        assert ast_from_value(Undefined, GraphQLString) is None

    def converts_using_serialize_from_a_custom_scalar_type():
        pass_through_scalar = GraphQLScalarType(
            "PassThroughScalar",
            serialize=lambda value: value,
        )

        assert ast_from_value("value", pass_through_scalar) == StringValueNode(
            value="value"
        )

        with raises(TypeError) as exc_info:
            assert ast_from_value(nan, pass_through_scalar)
        assert str(exc_info.value) == "Cannot convert value to AST: nan."

        with raises(TypeError) as exc_info:
            ast_from_value(inf, pass_through_scalar)
        assert str(exc_info.value) == "Cannot convert value to AST: inf."

        return_null_scalar = GraphQLScalarType(
            "ReturnNullScalar",
            serialize=lambda value: None,
        )

        assert ast_from_value("value", return_null_scalar) is None

        class SomeClass:
            pass

        return_custom_class_scalar = GraphQLScalarType(
            "ReturnCustomClassScalar",
            serialize=lambda value: SomeClass(),
        )

        with raises(TypeError) as exc_info:
            ast_from_value("value", return_custom_class_scalar)
        msg = str(exc_info.value)
        assert msg == "Cannot convert value to AST: <SomeClass instance>."

    def does_not_convert_non_null_values_to_null_value():
        non_null_boolean = GraphQLNonNull(GraphQLBoolean)
        assert ast_from_value(None, non_null_boolean) is None

    complex_value = {"someArbitrary": "complexValue"}

    my_enum = GraphQLEnumType(
        "MyEnum", {"HELLO": None, "GOODBYE": None, "COMPLEX": complex_value}
    )

    def converts_string_values_to_enum_asts_if_possible():
        assert ast_from_value("HELLO", my_enum) == EnumValueNode(value="HELLO")

        assert ast_from_value(complex_value, my_enum) == EnumValueNode(value="COMPLEX")

        # Note: case sensitive
        with raises(GraphQLError) as exc_info:
            ast_from_value("hello", my_enum)
        assert exc_info.value.message == "Enum 'MyEnum' cannot represent value: 'hello'"

        # Note: not a valid enum value
        with raises(GraphQLError) as exc_info:
            ast_from_value("UNKNOWN_VALUE", my_enum)
        assert (
            exc_info.value.message
            == "Enum 'MyEnum' cannot represent value: 'UNKNOWN_VALUE'"
        )

    def converts_list_values_to_list_asts():
        assert ast_from_value(
            ["FOO", "BAR"], GraphQLList(GraphQLString)
        ) == ListValueNode(
            values=[StringValueNode(value="FOO"), StringValueNode(value="BAR")]
        )

        assert ast_from_value(
            ["HELLO", "GOODBYE"], GraphQLList(my_enum)
        ) == ListValueNode(
            values=[EnumValueNode(value="HELLO"), EnumValueNode(value="GOODBYE")]
        )

        def list_generator():
            yield 1
            yield 2
            yield 3

        assert ast_from_value(list_generator(), GraphQLList(GraphQLInt)) == (
            ListValueNode(
                values=[
                    IntValueNode(value="1"),
                    IntValueNode(value="2"),
                    IntValueNode(value="3"),
                ]
            )
        )

    def converts_list_singletons():
        assert ast_from_value("FOO", GraphQLList(GraphQLString)) == StringValueNode(
            value="FOO"
        )

    def skips_invalid_list_items():
        ast = ast_from_value(
            ["FOO", None, "BAR"], GraphQLList(GraphQLNonNull(GraphQLString))
        )

        assert ast == ListValueNode(
            values=[StringValueNode(value="FOO"), StringValueNode(value="BAR")]
        )

    input_obj = GraphQLInputObjectType(
        "MyInputObj",
        {"foo": GraphQLInputField(GraphQLFloat), "bar": GraphQLInputField(my_enum)},
    )

    def converts_input_objects():
        assert ast_from_value({"foo": 3, "bar": "HELLO"}, input_obj) == ObjectValueNode(
            fields=[
                ObjectFieldNode(
                    name=NameNode(value="foo"), value=FloatValueNode(value="3")
                ),
                ObjectFieldNode(
                    name=NameNode(value="bar"), value=EnumValueNode(value="HELLO")
                ),
            ]
        )

    def converts_input_objects_with_explicit_nulls():
        assert ast_from_value({"foo": None}, input_obj) == ObjectValueNode(
            fields=[ObjectFieldNode(name=NameNode(value="foo"), value=NullValueNode())]
        )

    def does_not_convert_non_object_values_as_input_objects():
        assert ast_from_value(5, input_obj) is None



from operator import attrgetter
from typing import Any, Collection

from ...error import GraphQLError
from ...language import (
    DirectiveDefinitionNode,
    FieldDefinitionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    InterfaceTypeExtensionNode,
    NameNode,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
    VisitorAction,
    SKIP,
)
from ...pyutils import group_by
from . import SDLValidationRule

__all__ = ["UniqueArgumentDefinitionNamesRule"]


class UniqueArgumentDefinitionNamesRule(SDLValidationRule):
    """Unique argument definition names

    A GraphQL Object or Interface type is only valid if all its fields have uniquely
    named arguments.
    A GraphQL Directive is only valid if all its arguments are uniquely named.

    See https://spec.graphql.org/draft/#sec-Argument-Uniqueness
    """

    def enter_directive_definition(
        self, node: DirectiveDefinitionNode, *_args: Any
    ) -> VisitorAction:
        return self.check_arg_uniqueness(f"@{node.name.value}", node.arguments)

    def enter_interface_type_definition(
        self, node: InterfaceTypeDefinitionNode, *_args: Any
    ) -> VisitorAction:
        return self.check_arg_uniqueness_per_field(node.name, node.fields)

    def enter_interface_type_extension(
        self, node: InterfaceTypeExtensionNode, *_args: Any
    ) -> VisitorAction:
        return self.check_arg_uniqueness_per_field(node.name, node.fields)

    def enter_object_type_definition(
        self, node: ObjectTypeDefinitionNode, *_args: Any
    ) -> VisitorAction:
        return self.check_arg_uniqueness_per_field(node.name, node.fields)

    def enter_object_type_extension(
        self, node: ObjectTypeExtensionNode, *_args: Any
    ) -> VisitorAction:
        return self.check_arg_uniqueness_per_field(node.name, node.fields)

    def check_arg_uniqueness_per_field(
        self,
        name: NameNode,
        fields: Collection[FieldDefinitionNode],
    ) -> VisitorAction:
        type_name = name.value
        for field_def in fields:
            field_name = field_def.name.value
            argument_nodes = field_def.arguments or ()
            self.check_arg_uniqueness(f"{type_name}.{field_name}", argument_nodes)
        return SKIP

    def check_arg_uniqueness(
        self, parent_name: str, argument_nodes: Collection[InputValueDefinitionNode]
    ) -> VisitorAction:
        seen_args = group_by(argument_nodes, attrgetter("name.value"))
        for arg_name, arg_nodes in seen_args.items():
            if len(arg_nodes) > 1:
                self.report_error(
                    GraphQLError(
                        f"Argument '{parent_name}({arg_name}:)'"
                        " can only be defined once.",
                        [node.name for node in arg_nodes],
                    )
                )
        return SKIP
